@@ -41,9 +41,23 @@ namespace gr
 //:>********************************************************************************************
 //:>	Font methods
 //:>********************************************************************************************
+
+/*----------------------------------------------------------------------------------------------
+	Constructor.
+----------------------------------------------------------------------------------------------*/
+Font::Font(const Font & fontSrc) : m_pfface(fontSrc.m_pfface)
+{
+	if (m_pfface)
+		m_pfface->IncFontCount();
+}
+
+/*----------------------------------------------------------------------------------------------
+	Destructor.
+----------------------------------------------------------------------------------------------*/
 Font::~Font()
 {
-	if (m_pfface) m_pfface->DecFontCount();
+	if (m_pfface)
+		m_pfface->DecFontCount();
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -62,6 +76,40 @@ int Font::GetFlushMode()
 	return FontFace::GetFlushMode();
 }
 
+/*----------------------------------------------------------------------------------------------
+	Return the internal object representing the font-face. This is the object that is stored
+	in Graphite's internal cache.
+----------------------------------------------------------------------------------------------*/
+inline FontFace & Font::fontFace(bool fDumbFallback)
+{
+	if (m_pfface == 0)
+		initialiseFontFace(fDumbFallback);
+	return *m_pfface;
+}
+
+void Font::initialiseFontFace(bool fDumbFallback)
+{
+	std::wstring stuFaceName;
+	bool         fBold, fItalic;
+
+	UniqueCacheInfo(stuFaceName, fBold, fItalic);
+	
+	m_pfface = FontFace::GetFontFace(this, stuFaceName, fBold, fItalic, fDumbFallback);
+
+	Assert(m_pfface != 0);
+	
+	m_pfface->IncFontCount();
+
+	FontException fexptn;
+	fexptn.version = -1;
+	fexptn.subVersion = -1;
+	if (m_pfface->BadFont(&fexptn.errorCode)
+		|| (!fDumbFallback && m_pfface->DumbFallback(&fexptn.errorCode)))
+	{
+		throw fexptn;
+	}
+}
+
 /*----------------------------------------------------------------------------------------------
 	Return uniquely identifying information that will be used a the key for this font
 	in the font cache. This includes the font face name and the bold and italic flags.
@@ -69,8 +117,8 @@ int Font::GetFlushMode()
 void Font::UniqueCacheInfo(std::wstring & stuFace, bool & fBold, bool & fItalic)
 {
 	size_t cbSize;
-	const void * pNameTbl = getTable(TtfUtil::TableIdTag(ktiName), &cbSize);
-	long lOffset, lSize;
+	const byte * pNameTbl = static_cast<const byte *>(getTable(TtfUtil::TableIdTag(ktiName), &cbSize));
+	size_t lOffset, lSize;
 	if (!TtfUtil::Get31EngFamilyInfo(pNameTbl, lOffset, lSize))
 	{
 		// TODO: try to find any name in any arbitrary language.
@@ -79,16 +127,12 @@ void Font::UniqueCacheInfo(std::wstring & stuFace, bool & fBold, bool & fItalic)
 	}
 	// byte * pvName = (byte *)pNameTbl + lOffset;
 	utf16 rgchwFace[128];
-	int cchw = (lSize / isizeof(utf16)) + 1;
-	cchw = min(cchw, 128);
-	memcpy(rgchwFace, (byte *)pNameTbl + lOffset, lSize);
-	rgchwFace[cchw - 1] = 0;  // zero terminate
-	TtfUtil::SwapWString(rgchwFace, cchw - 1);
-	#if SIZEOF_WCHAR_T==2
-	stuFace.assign(rgchwFace);
-	#else
-	//#warning "need to fix Font::UniqueCacheInfo someday when it is used"
-	#endif
+	const size_t cchw = min(long(lSize / sizeof(utf16)), long(sizeof rgchwFace - 1));
+	const utf16 *src_start = reinterpret_cast<const utf16 *>(pNameTbl+ lOffset);
+	std::copy(src_start, src_start + cchw, rgchwFace);
+	rgchwFace[cchw] = 0;  // zero terminate
+	TtfUtil::SwapWString(rgchwFace, cchw);
+	stuFace.assign(rgchwFace, rgchwFace + cchw);
 
 	const void * pOs2Tbl = getTable(TtfUtil::TableIdTag(ktiOs2), &cbSize);
 	TtfUtil::FontOs2Style(pOs2Tbl, fBold, fItalic);
@@ -117,15 +161,35 @@ void Font::getGlyphPoint(gid16 glyphID, unsigned int pointNum, Point & pointRetu
 	const void * pLoca = getTable(TtfUtil::TableIdTag(ktiLoca), &cbLocaSize);
 	if (pLoca == 0)	return;
 
-	const size_t MAX_CONTOURS = 32;
-	int cnPoints = MAX_CONTOURS;
-	bool rgfOnCurve[MAX_CONTOURS];
-	int prgnX[MAX_CONTOURS];
-	int prgnY[MAX_CONTOURS];
-	
-	if (TtfUtil::GlyfPoints(glyphID, pGlyf, pLoca, 
-		cbLocaSize, pHead, 0, 0, 
-		prgnX, prgnY, rgfOnCurve, cnPoints))
+	size_t cContours;
+	size_t cPoints;
+
+	const size_t CONTOUR_BUF_SIZE = 16;
+	const size_t POINT_BUF_SIZE = 64;
+	// Fixed-size buffers that will be used in most cases:
+	bool rgfOnCurveFixedBuf[POINT_BUF_SIZE], *rgfOnCurveHeapBuf=0;
+	int  rgnEndPtFixedBuf[CONTOUR_BUF_SIZE], *rgnEndPtHeapBuf=0,
+	     rgnXFixedBuf[POINT_BUF_SIZE],       *rgnXHeapBuf=0,
+	     rgnYFixedBuf[POINT_BUF_SIZE],       *rgnYHeapBuf=0;
+
+	if (!TtfUtil::GlyfContourCount(glyphID, pGlyf, pLoca, cbLocaSize, pHead, cContours))
+		return;	//  can't figure it out
+
+	int * prgnEndPt = (cContours > CONTOUR_BUF_SIZE) ? rgnEndPtHeapBuf=new int[cContours] : rgnEndPtFixedBuf;
+
+	if (!TtfUtil::GlyfContourEndPoints(glyphID, pGlyf, pLoca, cbLocaSize, pHead, 
+		prgnEndPt, cContours))
+	{
+		return;	// should have been caught above
+	}
+	cPoints = prgnEndPt[cContours - 1] + 1;
+
+	bool * prgfOnCurve = (cPoints > POINT_BUF_SIZE) ? rgfOnCurveHeapBuf=new bool[cPoints] : rgfOnCurveFixedBuf;
+	int * prgnX        = (cPoints > POINT_BUF_SIZE) ? rgnXHeapBuf=new int[cPoints]  : rgnXFixedBuf;
+	int * prgnY        = (cPoints > POINT_BUF_SIZE) ? rgnYHeapBuf=new int[cPoints]  : rgnYFixedBuf;
+
+	if (TtfUtil::GlyfPoints(glyphID, pGlyf, pLoca, cbLocaSize, pHead, 0, 0, 
+		prgnX, prgnY, prgfOnCurve, cPoints))
 	{
 		float nPixEmSquare;
 		getFontMetrics(0, 0, &nPixEmSquare);
@@ -134,6 +198,11 @@ void Font::getGlyphPoint(gid16 glyphID, unsigned int pointNum, Point & pointRetu
 		pointReturn.x = prgnX[pointNum] / nDesignUnitsPerPixel;
 		pointReturn.y = prgnY[pointNum] / nDesignUnitsPerPixel;
 	}
+
+	delete[] rgnEndPtHeapBuf;
+	delete[] rgfOnCurveHeapBuf;
+	delete[] rgnXHeapBuf;
+	delete[] rgnYHeapBuf;
 }
 
 
@@ -165,7 +234,8 @@ void Font::getGlyphMetrics(gid16 glyphID, gr::Rect & boundingBox, gr::Point & ad
 		float(TtfUtil::DesignUnits(pHead)) / pixelEmSquare;
 
 	// Use the Hmtx and Head tables to find the glyph advances.
-	int lsb, advance = 0;
+	int lsb;
+	unsigned int advance = 0;
 	if (TtfUtil::HorMetrics(glyphID, pHmtx, hmtxSize, pHead, 
 			lsb, advance))
 	{
@@ -176,16 +246,18 @@ void Font::getGlyphMetrics(gid16 glyphID, gr::Rect & boundingBox, gr::Point & ad
 	const void * pGlyf = getTable(TtfUtil::TableIdTag(ktiGlyf), &locaSize);
 	if (pGlyf == 0)	return;
 
-	const void * pHhea = getTable(TtfUtil::TableIdTag(ktiHhea), &locaSize);
-	if (pHhea == 0)	return;
+//	const void * pHhea = getTable(TtfUtil::TableIdTag(ktiHhea), &locaSize);
+//	if (pHhea == 0)	return;
 
 	const void * pLoca = getTable(TtfUtil::TableIdTag(ktiLoca), &locaSize);
 	if (pLoca == 0)	return;
 
 	// Fetch the glyph bounding box, GlyphBox may return false for a 
-	//    whitespace glyph.
+	// whitespace glyph.
+	// Note that using GlyfBox here allows attachment points (ie, points lying outside
+	// the glyph's outline) to affect the bounding box, which might not be what we want.
 	int xMin, xMax, yMin, yMax;
-	if (TtfUtil::GlyfBox(glyphID, pGlyf, pLoca, locaSize, pHhea,
+	if (TtfUtil::GlyfBox(glyphID, pGlyf, pLoca, locaSize, pHead,
 			xMin, yMin, xMax, yMax))
 	{
 		boundingBox.left = (xMin / designUnitsPerPixel);
@@ -270,47 +342,67 @@ bool Font::getFeatureSettingLabel(FeatureSettingIterator fsit, lgid nLanguage, u
 ----------------------------------------------------------------------------------------------*/
 size_t Font::NumberOfFeatures()
 {
-	return m_pfface->NumberOfFeatures();
+	return fontFace().NumberOfFeatures();
 }
 featid Font::FeatureID(size_t ifeat)
 {
-	return m_pfface->FeatureID(ifeat);
+	return fontFace().FeatureID(ifeat);
 }
 size_t Font::FeatureWithID(featid id)
 {
-	return m_pfface->FeatureWithID(id);
+	return fontFace().FeatureWithID(id);
 }
 bool Font::GetFeatureLabel(size_t ifeat, lgid language, utf16 * label)
 {
-	return m_pfface->GetFeatureLabel(ifeat, language, label);
+	return fontFace().GetFeatureLabel(ifeat, language, label);
 }
 int Font::GetFeatureDefault(size_t ifeat) // index of default setting
 {
-	return m_pfface->GetFeatureDefault(ifeat);
+	return fontFace().GetFeatureDefault(ifeat);
 }
 size_t Font::NumberOfSettings(size_t ifeat)
 {
-	return m_pfface->NumberOfSettings(ifeat);
+	return fontFace().NumberOfSettings(ifeat);
 }
 int Font::GetFeatureSettingValue(size_t ifeat, size_t ifset)
 {
-	return m_pfface->GetFeatureSettingValue(ifeat, ifset);
+	return fontFace().GetFeatureSettingValue(ifeat, ifset);
 }
 bool Font::GetFeatureSettingLabel(size_t ifeat, size_t ifset, lgid language, utf16 * label)
 {
-	return m_pfface->GetFeatureSettingLabel(ifeat, ifset, language, label);
+	return fontFace().GetFeatureSettingLabel(ifeat, ifset, language, label);
+}
+size_t Font::NumberOfFeatLangs()
+{
+	return fontFace().NumberOfFeatLangs();
+}
+short Font::FeatLabelLang(size_t ilang)
+{
+	return fontFace().FeatLabelLang(ilang);
 }
 size_t Font::NumberOfLanguages()
 {
-	return m_pfface->NumberOfLanguages();
+	return fontFace().NumberOfLanguages();
 }
 isocode Font::LanguageCode(size_t ilang)
 {
-	return m_pfface->LanguageCode(ilang);
+	return fontFace().LanguageCode(ilang);
 }
 
 /*----------------------------------------------------------------------------------------------
-	Return a pair of iterators over a sequence of features defined for the font.
+	Return a pair of iterators over a sequence of feature-label languages for this font.
+----------------------------------------------------------------------------------------------*/
+std::pair<FeatLabelLangIterator, FeatLabelLangIterator> Font::getFeatureLabelLanguages()
+{
+	std::pair<FeatLabelLangIterator, FeatLabelLangIterator> pairRet;
+	pairRet.first = BeginFeatLang();
+	pairRet.second = EndFeatLang();
+	return pairRet;
+}
+
+/*----------------------------------------------------------------------------------------------
+	Return a pair of iterators over a sequence of languages defined for the font, that is,
+	those that have features specified for them.
 ----------------------------------------------------------------------------------------------*/
 std::pair<LanguageIterator, LanguageIterator> Font::getSupportedLanguages()
 {
@@ -320,28 +412,36 @@ std::pair<LanguageIterator, LanguageIterator> Font::getSupportedLanguages()
 	return pairRet;
 }
 
+/*----------------------------------------------------------------------------------------------
+	Return a bitmap indicating which directions are supported by the font.
+----------------------------------------------------------------------------------------------*/
+ScriptDirCode Font::getSupportedScriptDirections() const throw()
+{
+	return m_pfface->ScriptDirection();
+}
+
 /*----------------------------------------------------------------------------------------------
 	For segment creation.
 ----------------------------------------------------------------------------------------------*/
 void Font::RenderLineFillSegment(Segment * pseg, ITextSource * pts, LayoutEnvironment & layout,
 	toffset ichStart, toffset ichStop, float xsMaxWidth, bool fBacktracking)
 {
-	m_pfface->RenderLineFillSegment(pseg, this, pts, layout, ichStart, ichStop, xsMaxWidth,
-		fBacktracking);
+	fontFace(layout.dumbFallback()).RenderLineFillSegment(pseg, this, pts, layout,
+		ichStart, ichStop, xsMaxWidth, fBacktracking);
 }
 
 void Font::RenderRangeSegment(Segment * pseg, ITextSource * pts, LayoutEnvironment & layout,
 	toffset ichStart, toffset ichStop)
 {
-	m_pfface->RenderRangeSegment(pseg, this, pts, layout, ichStart, ichStop);
+	fontFace(layout.dumbFallback()).RenderRangeSegment(pseg, this, pts, layout, ichStart, ichStop);
 }
 
 void Font::RenderJustifiedSegment(Segment * pseg, ITextSource * pts,
 	LayoutEnvironment & layout, toffset ichStart, toffset ichStop,
 	float xsCurrentWidth, float xsDesiredWidth)
 {
-	m_pfface->RenderJustifiedSegment(pseg, this, pts, layout, ichStart, ichStop,
-		xsCurrentWidth, xsDesiredWidth);
+	fontFace(layout.dumbFallback()).RenderJustifiedSegment(pseg, this, pts, layout,
+		ichStart, ichStop, xsCurrentWidth, xsDesiredWidth);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -360,6 +460,22 @@ FeatureIterator Font::EndFeature()
 	return fit;
 }
 
+/*----------------------------------------------------------------------------------------------
+	Feature-label language iterators.
+----------------------------------------------------------------------------------------------*/
+FeatLabelLangIterator Font::BeginFeatLang()
+{
+	FeatLabelLangIterator lgit(this, 0, NumberOfFeatLangs());
+	return lgit;
+}
+
+FeatLabelLangIterator Font::EndFeatLang()
+{
+	int clang = NumberOfFeatLangs();
+	FeatLabelLangIterator lgit(this, clang, clang);
+	return lgit;
+}
+
 
 /*----------------------------------------------------------------------------------------------
 	Language iterators.
@@ -378,6 +494,15 @@ LanguageIterator Font::EndLanguage()
 }
 
 
+/*----------------------------------------------------------------------------------------------
+	Debugging.
+----------------------------------------------------------------------------------------------*/
+//bool Font::DbgCheckFontCache()
+//{
+//	return FontFace::DbgCheckFontCache();
+//}
+
+
 //:>********************************************************************************************
 //:>	FeatureIterator methods
 //:>********************************************************************************************
@@ -397,7 +522,7 @@ featid FeatureIterator::operator*()
 	if (m_ifeat >= m_cfeat)
 	{
 		Assert(false);
-		return kInvalid;
+		return (unsigned int)kInvalid;
 	}
 
 	return m_pfont->FeatureID(m_ifeat);
@@ -579,6 +704,89 @@ int FeatureSettingIterator::operator-(FeatureSettingIterator fsit)
 	return (m_ifset - fsit.m_ifset);
 }
 
+//:>********************************************************************************************
+//:>	FeatLabelLangIterator methods
+//:>********************************************************************************************
+
+/*----------------------------------------------------------------------------------------------
+	Dereference the iterator, returning a language LCID.
+----------------------------------------------------------------------------------------------*/
+data16 FeatLabelLangIterator::operator*()
+{
+	if (m_ilang >= m_clang)
+	{
+		Assert(false);
+		return 0;
+	}
+
+	return m_pfont->FeatLabelLang(m_ilang);
+}
+
+/*----------------------------------------------------------------------------------------------
+	Increment the iterator.
+----------------------------------------------------------------------------------------------*/
+FeatLabelLangIterator FeatLabelLangIterator::operator++()
+{
+	if (m_ilang >= m_clang)
+	{
+		// Can't increment.
+		Assert(false);
+	}
+	else
+		m_ilang++;
+
+	return *this;
+}
+
+/*----------------------------------------------------------------------------------------------
+	Increment the iterator by the given value.
+----------------------------------------------------------------------------------------------*/
+FeatLabelLangIterator FeatLabelLangIterator::operator+=(int n)
+{
+	if (m_ilang + n >= m_clang)
+	{
+		// Can't increment.
+		Assert(false);
+		m_ilang = m_clang;
+	}
+	else if (m_ilang + n < 0)
+	{
+		// Can't decrement.
+		Assert(false);
+		m_ilang = 0;
+	}
+	else
+		m_ilang += m_clang;
+
+	return *this;
+}
+
+/*----------------------------------------------------------------------------------------------
+	Test whether the two iterators are equal.
+----------------------------------------------------------------------------------------------*/
+bool FeatLabelLangIterator::operator==(FeatLabelLangIterator & fllit)
+{
+	return (fllit.m_ilang == m_ilang && fllit.m_pfont == m_pfont);
+}
+
+bool FeatLabelLangIterator::operator!=(FeatLabelLangIterator & fllit)
+{
+	return (fllit.m_ilang != m_ilang || fllit.m_pfont != m_pfont);
+}
+
+/*----------------------------------------------------------------------------------------------
+	Return the number of items represented by the range of the two iterators.
+----------------------------------------------------------------------------------------------*/
+int FeatLabelLangIterator::operator-(FeatLabelLangIterator & fllit)
+{
+	if (m_pfont != fllit.m_pfont)
+	{
+		throw;
+	}
+	return (m_ilang - fllit.m_ilang);
+}
+
+
 //:>********************************************************************************************
 //:>	LanguageIterator methods
 //:>********************************************************************************************
@@ -593,7 +801,7 @@ isocode LanguageIterator::operator*()
 		Assert(false);
 		isocode codeRet;
 		codeRet.rgch[1] = '?'; codeRet.rgch[2] = '?';
-		codeRet.rgch[3] = '?';  codeRet.rgch[2] = 0;
+		codeRet.rgch[3] = '?'; codeRet.rgch[4] = 0;
 		return codeRet;
 	}
 
