@@ -158,7 +158,13 @@ LgIpValidResult SegmentPainter::isValidInsertionPoint(int ichw)
 	int ichwSegOffset = ichw - m_pseg->m_ichwMin;
 	if (ichwSegOffset < m_pseg->m_ichwAssocsMin || ichwSegOffset >= m_pseg->m_ichwAssocsLim)
 	{
-		return kipvrUnknown;
+		// If we're pointing at a hard line break character (Unicode 2028), then we know
+		// that it's a good insertion point in spite of missing the other conditions.
+		utf16 ch;
+		m_pseg->m_pgts->fetch(ichw, 1, &ch);
+		if (ch != knLineSep)
+			return kipvrUnknown; // Definitely in a different segment of the paragraph.
+		return kipvrOK;
 	}
 
 	int isloutLig = m_pseg->m_prgisloutLigature[ichwSegOffset - m_pseg->m_ichwAssocsMin];
@@ -361,14 +367,14 @@ bool SegmentPainter::drawSelectionRange(int ichwAnchor, int ichwEnd,
 	std::vector<bool> vfEntireHt;
 
 	bool * prgfAllCompsSelected = new bool[m_pseg->m_dichwLim];
-	memset(prgfAllCompsSelected, 0, m_pseg->m_dichwLim);
+	std::fill_n(prgfAllCompsSelected, m_pseg->m_dichwLim, false);
 	CalcPartialLigatures(prgfAllCompsSelected, ichwMinSeg, ichwLimSeg,
 		ichwMinSel, ichwLimSel);
 
 	//	Array to keep track of which glyphs were highlighted.
 	int cginf = m_pseg->m_cginf;
 	bool * prgfHighlighted = new bool[cginf];
-	memset(prgfHighlighted, 0, cginf);
+	std::fill_n(prgfHighlighted, cginf, false);
 
 	for (int ichwLp = ichwMinSeg; ichwLp < ichwLimSeg; ichwLp++)
 	{
@@ -406,6 +412,13 @@ bool SegmentPainter::drawSelectionRange(int ichwAnchor, int ichwEnd,
 		rdTmp.left = min(xdSegRight, max(float(0), xdTmp));
 		xdTmp = ScaleXToDest(rsTmp.right + xsSegLeft);
 		rdTmp.right = max(float(0), min(xdSegRight, xdTmp));
+		if (rdTmp.left > rdTmp.right)
+		{
+			// I think this can happen due to rounding errors.
+			float tmp = rdTmp.left;
+			rdTmp.left = rdTmp.right;
+			rdTmp.right = tmp;
+		}
 		vrd.push_back(rdTmp);
 	}
 
@@ -478,12 +491,10 @@ bool SegmentPainter::positionsOfRange(int ichwAnchor, int ichwEnd,
 		prsBounds->right = max(prsBounds->right, prs->right);
 	}
 
-	int xsSegLeft = 0;
-
 	prdBounds->top = ydLineTop;
 	prdBounds->bottom = ydLineBottom;
-	prdBounds->left = ScaleXToDest(xsSegLeft + prsBounds->left);
-	prdBounds->right = ScaleXToDest(xsSegLeft + prsBounds->right);
+	prdBounds->left = ScaleXToDest(prsBounds->left);
+	prdBounds->right = ScaleXToDest(prsBounds->right);
 
 	return true;
 }
@@ -511,7 +522,7 @@ void SegmentPainter::pointToChar(Point zptdClickPosition, int * pichw, bool * pf
 
 	if (m_pseg->m_dichwLim == 0)
 	{
-		*pichw = 0;
+		*pichw = m_pseg->m_ichwMin;
 		*pfAssocPrev = false;
 		//RestoreFont(pgg);
 		return;
@@ -828,8 +839,6 @@ GrResult SegmentPainter::ArrowKeyPositionAux(
 /*----------------------------------------------------------------------------------------------
 	Used to find out where underlines should be drawn.
 
-	TODO SharonC: handle end-of-line spaces properly--they shouldn't be underlined.
-
 	@param ichwMin/Lim		- range of text of interest
 	@param rs, rd			- source/destination coordinates, for scaling
 	@param fSkipSpace		- true if white space should not be underlined; some renderers may
@@ -858,9 +867,14 @@ size_t SegmentPainter::getUnderlinePlacement(int ichwMin, int ichwLim,
 	Assert(m_pseg->m_dxsWidth >= 0);
 	Assert(m_pseg->m_dysAscent >= 0);
 
-	float xsSegLeft = 0;
+	float xsSegLeft = m_pseg->m_dxsTotalWidth;
+	for (int iginf = 0; iginf < m_pseg->m_cginf; iginf++)
+	{
+		xsSegLeft = min(xsSegLeft, m_pseg->GlyphLeftEdge(iginf));
+	}
 
-	float xdSegRight = this->ScaleXToDest(m_pseg->m_dxsTotalWidth);
+	float xdSegLeft = this->ScaleXToDest(xsSegLeft);
+	float xdSegRight = xdSegLeft + m_pseg->m_dxsTotalWidth;
 
 	int ichwMinSel = min(ichwMin, ichwLim);
 	int ichwLimSel = max(ichwMin, ichwLim);
@@ -872,14 +886,14 @@ size_t SegmentPainter::getUnderlinePlacement(int ichwMin, int ichwLim,
 	std::vector<bool> vfEntireHt;
 
 	bool * prgfAllCompsSelected = new bool[m_pseg->m_dichwLim];
-	memset(prgfAllCompsSelected, 0, m_pseg->m_dichwLim);
+	std::fill_n(prgfAllCompsSelected, m_pseg->m_dichwLim, false);
 	CalcPartialLigatures(prgfAllCompsSelected, ichwMinSeg, ichwLimSeg,
 		ichwMinSel, ichwLimSel);
 
 	//	Array to keep track of which glyphs were highlighted:
 	int cginf = m_pseg->m_cginf;
 	bool * prgfHighlighted = new bool[cginf];
-	memset(prgfHighlighted, 0, cginf);
+	std::fill_n(prgfHighlighted, cginf, false);
 
 	for (int ichwLp = ichwMinSeg; ichwLp < ichwLimSeg; ichwLp++)
 	{
@@ -902,8 +916,8 @@ size_t SegmentPainter::getUnderlinePlacement(int ichwMin, int ichwLim,
 	{
 		Rect rsTmp = vrs[irs];
 		LineSeg lsdTmp;
-		lsdTmp.left = min(xdSegRight, max(float(0), ScaleXToDest(rsTmp.left + xsSegLeft)));
-		lsdTmp.right = max(float(0), min(xdSegRight, ScaleXToDest(rsTmp.right + xsSegLeft)));
+		lsdTmp.left = min(xdSegRight, max(xdSegLeft, ScaleXToDest(rsTmp.left)));
+		lsdTmp.right = max(xdSegLeft, min(xdSegRight, ScaleXToDest(rsTmp.right)));
 		vlsd.push_back(lsdTmp);
 	}
 
@@ -1003,7 +1017,7 @@ void SegmentPainter::CalcOrDrawInsertionPoint(
 	CalcIP(ichwIP, true, &xsBefore, &ysBeforeTop, &ysBeforeBottom, &fRtlBefore);
 	CalcIP(ichwIP, false, &xsAfter, &ysAfterTop,  &ysAfterBottom,  &fRtlAfter);
 
-	if (m_pseg->m_cginf == 0)
+	if (m_pseg->m_cginf == 0 && ichwIP == ichwLim)
 	{
 		Assert(kPosInfFloat == xsBefore && kPosInfFloat == xsAfter);
 		xsBefore = 0;
@@ -1494,78 +1508,89 @@ void SegmentPainter::CalcIP(int ichwSel, bool fBefore,
 		if (kPosInfinity == iginfOutput || kNegInfinity == iginfOutput)
 			*pxs = kPosInfFloat;
 
-		else if (psloutTmp->IsPartOfCluster())
+		else
 		{
-			//	Cluster member.
-
-			if (AtEdgeOfCluster(psloutTmp, islout, fBefore) &&
-				!CanInsertIntoCluster(psloutTmp, islout))
+			bool fCluster = (psloutTmp->IsPartOfCluster());
+			bool fHighlightBB = fCluster;
+			// Test below should match the one in CalcHighlightRect:
+			if ((psloutTmp->ClusterAdvance() == 0 || psloutTmp->GlyphMetricLogUnits(kgmetAdvWidth) == 0)
+				&& !psloutTmp->IsSpace())
 			{
-				//	Place the IP at the edge of the whole cluster.
+				fHighlightBB = true;	// use bounding box for zero-advance glyph
+			}
+			if (fHighlightBB)
+			{
+				//	Cluster member or zero-width glyph.
 
-				int isloutBase = psloutTmp->ClusterBase();
-				GrSlotOutput * psloutBase = m_pseg->OutputSlot(isloutBase);
-				//int igbbBase = psloutBase->GlyphBbIndex();
-				int iginfBase = m_pseg->LogicalToPhysicalSurface(isloutBase);
-				
-				float xsBase = m_pseg->GlyphLeftEdge(iginfBase);
+				if (fCluster && AtEdgeOfCluster(psloutTmp, islout, fBefore) &&
+					!CanInsertIntoCluster(psloutTmp, islout))
+				{
+					//	Place the IP at the edge of the whole cluster.
 
-				float xsLeft = xsBase + psloutBase->ClusterXOffset();
-				float xsRight = xsBase + psloutBase->ClusterAdvance();
+					int isloutBase = psloutTmp->ClusterBase();
+					GrSlotOutput * psloutBase = m_pseg->OutputSlot(isloutBase);
+					//int igbbBase = psloutBase->GlyphBbIndex();
+					int iginfBase = m_pseg->LogicalToPhysicalSurface(isloutBase);
+					
+					float xsBase = m_pseg->GlyphLeftEdge(iginfBase);
 
-				if (fBefore == *pfRtl)
-					//	Position of interest is to the right
-					*pxs = xsRight;
+					float xsLeft = xsBase + psloutBase->ClusterXOffset();
+					float xsRight = xsBase + psloutBase->ClusterAdvance();
+
+					if (fBefore == *pfRtl)
+						//	Position of interest is to the right
+						*pxs = xsRight;
+					else
+						//	Position of interest is to the left.
+						*pxs = xsLeft;
+				}
 				else
-					//	Position of interest is to the left.
-					*pxs = xsLeft;
+				{
+					//	Place the IP on the perimeter of the bounding box of the cluster member.
+
+					float xsGlyphStart = m_pseg->GlyphLeftEdge(iginfOutput);
+
+//					float dysFontAscent;
+//					res = GetFontAscentSourceUnits(pgg, &dysFontAscent);
+//					if (ResultFailed(res))	THROW(WARN(res));
+//					dysFontAscent += m_dysXAscent;
+
+					float ysBbTop = psloutTmp->GlyphMetricLogUnits(kgmetBbTop);
+					float ysBbBottom = psloutTmp->GlyphMetricLogUnits(kgmetBbBottom);
+					float xsBbLeft = psloutTmp->GlyphMetricLogUnits(kgmetBbLeft);
+					float xsBbRight = psloutTmp->GlyphMetricLogUnits(kgmetBbRight);
+
+					int twoPixels = 2;	// assume dest coords are in pixels
+
+					if (fBefore == *pfRtl)
+						//	Position of interest is to the right
+						*pxs = xsGlyphStart + xsBbRight + twoPixels;
+					else
+						//	Position of interest is to the left.
+						*pxs = xsGlyphStart + xsBbLeft - twoPixels;
+
+					GlyphInfo * pginf = m_pseg->m_prgginf + iginfOutput;
+					float ysGlyph = pginf->yOffset();
+					float ysGlyphBaseline = m_pseg->m_dysAscent - ysGlyph;	// relative to top
+
+					*pysTop = ysGlyphBaseline - ysBbTop - twoPixels;
+					*pysBottom = ysGlyphBaseline - ysBbBottom + twoPixels;
+				}
 			}
 			else
 			{
-				//	Place the IP on the perimeter of the bounding box of the cluster member.
-
-				float xsGlyphStart = m_pseg->GlyphLeftEdge(iginfOutput);
-
-//				float dysFontAscent;
-//				res = GetFontAscentSourceUnits(pgg, &dysFontAscent);
-//				if (ResultFailed(res))	THROW(WARN(res));
-//				dysFontAscent += m_dysXAscent;
-
-				float ysBbTop = psloutTmp->GlyphMetricLogUnits(kgmetBbTop);
-				float ysBbBottom = psloutTmp->GlyphMetricLogUnits(kgmetBbBottom);
-				float xsBbLeft = psloutTmp->GlyphMetricLogUnits(kgmetBbLeft);
-				float xsBbRight = psloutTmp->GlyphMetricLogUnits(kgmetBbRight);
-
-				int twoPixels = 2;	// assume dest coords are in pixels
-
+				// Insertion point goes at origin or advance width of glyph.
+	
+				*pxs = m_pseg->GlyphLeftEdge(iginfOutput);
 				if (fBefore == *pfRtl)
+				{
 					//	Position of interest is to the right
-					*pxs = xsGlyphStart + xsBbRight + twoPixels;
-				else
-					//	Position of interest is to the left.
-					*pxs = xsGlyphStart + xsBbLeft - twoPixels;
-
-				GlyphInfo * pginf = m_pseg->m_prgginf + iginfOutput;
-				float ysGlyph = pginf->yOffset();
-				float ysGlyphBaseline = m_pseg->m_dysAscent - ysGlyph;	// relative to top
-
-				*pysTop = ysGlyphBaseline - ysBbTop - twoPixels;
-				*pysBottom = ysGlyphBaseline - ysBbBottom + twoPixels;
+					//*pxs += psloutTmp->GlyphMetricLogUnits(kgmetAdvWidth);
+					*pxs += psloutTmp->ClusterAdvance();
+				}
+				// else position of interest is to the left
 			}
 		}
-		else
-		{
-			// Insertion point goes at origin or advance width of glyph.
-
-			*pxs = m_pseg->GlyphLeftEdge(iginfOutput);
-			if (fBefore == *pfRtl)
-			{
-				//	Position of interest is to the right
-				//*pxs += psloutTmp->GlyphMetricLogUnits(kgmetAdvWidth);
-				*pxs += psloutTmp->ClusterAdvance();
-			}
-			// else position of interest is to the left
-		}
 	}
 }
 
@@ -1621,11 +1646,37 @@ bool SegmentPainter::AtEdgeOfCluster(GrSlotOutput * pslout, int islout, bool fBe
 
 	//	If the adjacent slot in whatever direction we're interested in has the same
 	//	cluster base, this slot is not on the edge of the cluster.
-	int isloutAdjacentBase = (fBefore) ?
-		m_pseg->OutputSlot(islout - 1)->ClusterBase() :
-		m_pseg->OutputSlot(islout + 1)->ClusterBase();
+	//  -- NO, because you could have an intervening glyph that is not part of the
+	//	cluster.
+	//int isloutAdjacentBase = (fBefore) ?
+	//	m_pseg->OutputSlot(islout - 1)->ClusterBase() :
+	//	m_pseg->OutputSlot(islout + 1)->ClusterBase();
+	//return (isloutAdjacentBase != m_pseg->OutputSlot(islout)->ClusterBase());
 
-	return (isloutAdjacentBase != m_pseg->OutputSlot(islout)->ClusterBase());
+	// Ask the cluster base.
+	return AtEdgeOfCluster(m_pseg->OutputSlot(pslout->ClusterBase()), pslout->ClusterBase(),
+		pslout, islout, fBefore);
+}
+
+bool SegmentPainter::AtEdgeOfCluster(GrSlotOutput * psloutBase, int isloutBase,
+	GrSlotOutput * pslout, int islout, bool fBefore)
+{
+	//	Compare pslout to all the members of the cluster. If it is the minimum or maximum, it
+	//	is at an edge.
+	if (fBefore && isloutBase < islout)
+		return false;
+	if (!fBefore && isloutBase > islout)
+		return false;
+
+	for (int iislout = 0; iislout < psloutBase->NumClusterMembers(); iislout++)
+	{
+		int isloutMember = psloutBase->ClusterMember(iislout);
+		if (fBefore && isloutMember < islout)
+			return false;
+		if (!fBefore && isloutMember > islout)
+			return false;
+	}
+	return true;
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -1806,8 +1857,17 @@ void SegmentPainter::CalcHighlightRect(int ichw,
 
 			Rect rsNew;
 			GrSlotOutput * psloutTmp = m_pseg->OutputSlot(islout);
+			bool fHighlightBB = psloutTmp->IsPartOfCluster();
+			//	Test below should match the one in CalcIP--
+			//	If the glyph has no advance width but does have a real bounding box,
+			//	highlight the bounding box. Review: do we want this???
+			if ((psloutTmp->ClusterAdvance() == 0 || psloutTmp->GlyphMetricLogUnits(kgmetAdvWidth) == 0)
+				&& !psloutTmp->IsSpace())
+			{
+				fHighlightBB = true;
+			}
 			xsGlyphLeft = m_pseg->GlyphLeftEdge(iginfTmp);
-			if (psloutTmp->IsPartOfCluster())
+			if (fHighlightBB)
 			{
 				//	For now, highlight just the bounding box of this glyph. After we finish 
 				//	highlighting all glyphs individually, we'll go back and highlight
@@ -1842,6 +1902,9 @@ void SegmentPainter::CalcHighlightRect(int ichw,
 				//xsGlyphRight = xsGlyphLeft + psloutTmp->GlyphMetricLogUnits(kgmetAdvWidth);
 				xsGlyphRight = xsGlyphLeft + psloutTmp->ClusterAdvance();
 
+				if (xsGlyphLeft > xsGlyphRight)
+					std::swap(xsGlyphLeft, xsGlyphRight);
+				
 				rsNew.top = 0;
 				rsNew.bottom = m_pseg->m_dysHeight;
 				rsNew.left = xsGlyphLeft;
