@@ -47,7 +47,7 @@ bool compareCmap(const byte *lhs, const byte *rhs)
     using namespace gr;
     
     typedef const struct index {data16 version; data16 numberSubtables;}   *const IndexPtr;
-    typedef const struct enc {data16 platID; data16 platSpecID; DWORD offset;} *const EncPtr;
+    typedef const struct enc {data16 platID; data16 platSpecID; data32 offset;} *const EncPtr;
 
     // Calculate the size of the cmap.
 
@@ -59,7 +59,7 @@ bool compareCmap(const byte *lhs, const byte *rhs)
     for (size_t iSubTbl = 0; iSubTbl < numSubTbl; iSubTbl++)
     //for (int iSubTbl = numSubTbl - 1; iSubTbl >= 0; --iSubTbl)
     {
-        const WORD *const mapTbl = reinterpret_cast<const WORD*>(rhs + lsbf((int)encTbl[iSubTbl].offset));
+        const data16 *const mapTbl = reinterpret_cast<const data16 *>(rhs + lsbf((int)encTbl[iSubTbl].offset));
 		int format = lsbf(*mapTbl);
         switch (format)
         {
@@ -74,7 +74,7 @@ bool compareCmap(const byte *lhs, const byte *rhs)
             case 8:
 			case 10:
 			case 12:
-                cbCmapSz += lsbf((int)unsigned(reinterpret_cast<const DWORD *>(mapTbl)[1]));
+                cbCmapSz += lsbf((int)unsigned(reinterpret_cast<const data32 *>(mapTbl)[1]));
                 break;
 			default:
 				Assert(false);
@@ -119,6 +119,8 @@ void GrEngine::BasicInit()
 	// new interface stuff:
 	m_pCmapTbl = NULL;
 	m_pNameTbl = NULL;
+	m_fCmapTblCopy = false;
+	m_fNameTblCopy = false;
 
 	m_fLogXductn = false;
 
@@ -148,6 +150,9 @@ void GrEngine::BasicInit()
 	m_nFontCheckSum = 0;
 
 	m_fInErrorState = false;
+
+	m_rglcidFeatLabelLangs = NULL;	// initialize lazily, when needed
+	m_clcidFeatLabelLangs = 0;
 }
 
 GrEngine::~GrEngine()
@@ -200,10 +205,14 @@ void GrEngine::DestroyContents(bool fDestroyCmap)
 		m_pCmap_3_1 = NULL;
 		m_pCmap_3_10 = NULL;
 
-		delete[] m_pCmapTbl;
-		delete[] m_pNameTbl;
+		if (m_fCmapTblCopy)
+			delete[] m_pCmapTbl;
+		if (m_fNameTblCopy)
+			delete[] m_pNameTbl;
 		m_pCmapTbl = NULL;
 		m_pNameTbl = NULL;
+		m_fCmapTblCopy = false;
+		m_fNameTblCopy = false;
 	}
 
 	delete m_ptman;
@@ -260,8 +269,7 @@ bool GrEngine::GetFeatureLabel_ff(size_t ifeat, lgid nLanguage, utf16 * rgchwLab
 
 	int cch = stu.size();
 	cch = min(cch, 127); // 1 char for zero-termination
-
-	memcpy(rgchwLabel, stu.data(), (cch * isizeof(OLECHAR)));
+	std::copy(stu.data(), stu.data() + cch, rgchwLabel);
 	rgchwLabel[cch] = 0;
 
 	return (cch > 0);
@@ -312,13 +320,32 @@ bool GrEngine::GetFeatureSettingLabel_ff(size_t ifeat, size_t ifset, lgid langua
 
 	int cch = stu.size();
 	cch = min(cch, 127); // 1 char for zero-termination
-
-	memcpy(rgchwLabel, stu.data(), (cch * isizeof(OLECHAR)));
+	// Note: the wchar_t label was originally assigned from utf16 data, so although wchar_t is 
+	// utf32 on some platforms the conversion back to utf16 should still give correct results.
+	std::copy(stu.data(), stu.data() + cch, rgchwLabel);
 	rgchwLabel[cch] = 0;
 
 	return (cch > 0);
 }
 
+/*----------------------------------------------------------------------------------------------
+	For FontFace: return the number of languages that are possible among the feature labels.
+----------------------------------------------------------------------------------------------*/
+size_t GrEngine::NumberOfFeatLangs_ff()
+{
+	SetUpFeatLangList();
+	return m_clcidFeatLabelLangs;
+}
+
+/*----------------------------------------------------------------------------------------------
+	For FontFace: return the language LCID for the feature-label language with the given index.
+----------------------------------------------------------------------------------------------*/
+short GrEngine::GetFeatLabelLang_ff(size_t ilang)
+{
+	SetUpFeatLangList();
+	return m_rglcidFeatLabelLangs[ilang];
+}
+
 /*----------------------------------------------------------------------------------------------
 	For FontFace: return the number of supported languages.
 ----------------------------------------------------------------------------------------------*/
@@ -335,6 +362,23 @@ isocode GrEngine::GetLanguageCode_ff(size_t ilang)
 	return m_langtbl.LanguageCode(ilang);
 }
 
+/*----------------------------------------------------------------------------------------------
+	Set up the list of all the languages that are present in the feature labels.
+----------------------------------------------------------------------------------------------*/
+void GrEngine::SetUpFeatLangList()
+{
+	if (m_rglcidFeatLabelLangs)
+		return;
+
+	int rgnNameIDs[kMaxFeatures];
+	for (int ifeat = 0; ifeat < m_cfeat; ifeat++)
+		rgnNameIDs[ifeat] = m_rgfeat[ifeat].NameId();
+	short rglcid[128]; // 128 is the number expected by TtfUtil::GetLangsForNames
+	m_clcidFeatLabelLangs = TtfUtil::GetLangsForNames(m_pNameTbl, 3, 1, rgnNameIDs, m_cfeat, rglcid);
+	m_rglcidFeatLabelLangs = new short[m_clcidFeatLabelLangs];
+	memcpy(m_rglcidFeatLabelLangs, rglcid, sizeof(short) * m_clcidFeatLabelLangs);
+}
+
 /*----------------------------------------------------------------------------------------------
 	Return the maximum size needed for the block of data to pass between segments, that is,
 	to reinitialize the engine based on the results of the previously generated segment.
@@ -372,7 +416,7 @@ GrResult GrEngine::get_SegDatMaxLength(int * pcb)
 	@return The supported script direction(s). If more than one, the application is
 	responsible for choosing the most appropriate.
 ----------------------------------------------------------------------------------------------*/
-GrResult GrEngine::get_ScriptDirection(int * pgrfsdc, OLECHAR * prgchwErrMsg, int cchMaxErrMsg)
+GrResult GrEngine::get_ScriptDirection(unsigned int * pgrfsdc, OLECHAR * prgchwErrMsg, int cchMaxErrMsg)
 {
 	ChkGrOutPtr(pgrfsdc);
 	ChkGrArrayArg(prgchwErrMsg, cchMaxErrMsg);
@@ -465,7 +509,7 @@ GrResult GrEngine::MakeMeasuredSegment(
 		prgchwErrMsg, cchMaxErrMsg,
 		-1, false, false, klbClipBreak,	// not used
 		&dichLimSeg, pest,
-		100000, kestMoreLines);			// not used
+		false, 100000, kestMoreLines);			// not used
 }
 */
 
@@ -521,7 +565,7 @@ void GrEngine::MakeSegment(
 	float dxMaxWidth,
 	bool fBacktracking,
 	// for justification
-	float dxJustifiedWidthJ, SegEnd estJ)
+	bool fJust, float dxJustifiedWidthJ, SegEnd estJ)
 {
 	ChkGrArgPtr(pgts);
 	ChkGrArgPtrN(pjus);
@@ -535,8 +579,10 @@ void GrEngine::MakeSegment(
 		// Uninitialized.
 		return;   // ReturnResult(kresUnexpected);
 
+	SetCmapAndNameTables(pfont);
+
 	int segmode = ksegmodeRange; // segment from range
-	if (dxJustifiedWidthJ > 0)
+	if (fJust)
 		segmode = ksegmodeJust;	// justified segment
 	else if (dxMaxWidth < kPosInfFloat)
 		segmode = ksegmodeBreak;	// find break point
@@ -559,7 +605,9 @@ void GrEngine::MakeSegment(
 
 	//	Find the end of the range to render with the current font.
 	int nDirDepth;
-	int ichFontLim, ichStrmLim, ichSegLim;
+	int ichFontLim;
+    int ichStrmLim = 0;
+    int ichSegLim = 0;
 	switch (segmode)
 	{
 	case ksegmodeBreak:	// find break point
@@ -615,9 +663,9 @@ void GrEngine::MakeSegment(
 
 	//	Run the tables and get back a segment.
 
-	bool fMoreText;
-	int ichCallerBtLim;
-	bool fInfiniteWidth;
+	bool fMoreText = false;
+	int ichCallerBtLim = -1;
+	bool fInfiniteWidth = false;
 	switch (segmode)
 	{
 	//case ksegmodeMms:
@@ -737,9 +785,9 @@ void GrEngine::RecordFontLoadError(OLECHAR * prgchwErrMsg, int cchMax)
 		stuMessage.append(m_stuInitError);
 	}
 
-	memset(prgchwErrMsg, 0, isizeof(OLECHAR) * cchMax);
-	memcpy(prgchwErrMsg, stuMessage.data(),
-		min(cchMax, signed(stuMessage.size())) * isizeof(OLECHAR));
+	std::fill_n(prgchwErrMsg, cchMax, 0);
+	std::copy(stuMessage.data(), stuMessage.data() + min(cchMax - 1, signed(stuMessage.size())),
+				prgchwErrMsg);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -762,9 +810,9 @@ void GrEngine::RecordFontRunError(OLECHAR * prgchwErrMsg, int cchMax, GrResult r
 	else
 		stuMessage.append(L"\"");
 
-	memset(prgchwErrMsg, 0, isizeof(OLECHAR) * cchMax);
-	memcpy(prgchwErrMsg, stuMessage.data(),
-		min(cchMax, signed(stuMessage.size())) * isizeof(OLECHAR));
+	std::fill_n(prgchwErrMsg, cchMax, 0);
+	std::copy(stuMessage.data(), stuMessage.data() + min(cchMax - 1, signed(stuMessage.size())),
+			 prgchwErrMsg);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -774,7 +822,7 @@ void GrEngine::RecordFontRunError(OLECHAR * prgchwErrMsg, int cchMax, GrResult r
 ----------------------------------------------------------------------------------------------*/
 void GrEngine::ClearFontError(OLECHAR * prgchwErrMsg, int cchMaxErrMsg)
 {
-	memset(prgchwErrMsg, 0, isizeof(OLECHAR) * cchMaxErrMsg);
+	std::fill_n(prgchwErrMsg, cchMaxErrMsg, 0);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -848,26 +896,26 @@ int GrEngine::FindFontLim(ITextSource * pgts, int ichwMinFont, int * pnDirDepth)
 	implementation of the engine can't handle.
 ----------------------------------------------------------------------------------------------*/
 bool GrEngine::CheckTableVersions(GrIStream * pgrstrm,
-		byte *silf_tbl, int lSilfStart,
-		byte *gloc_tbl, int lGlocStart,
-		byte *feat_tbl, int lFeatStart,
+		byte *pSilfTbl, int lSilfStart,
+		byte *pGlobTbl, int lGlocStart,
+		byte *pFeatTbl, int lFeatStart,
 		int * pfxdBadVersion)
 {
-	pgrstrm->OpenBuffer(silf_tbl, isizeof(int));
+	pgrstrm->OpenBuffer(pSilfTbl, isizeof(int));
     pgrstrm->SetPositionInFont(lSilfStart);
 	*pfxdBadVersion = ReadVersion(*pgrstrm);
 	pgrstrm->CloseBuffer();
 	if (*pfxdBadVersion > kSilfVersion)
 		return false;
 
-	pgrstrm->OpenBuffer(gloc_tbl, lGlocStart + isizeof(int));
+	pgrstrm->OpenBuffer(pGlobTbl, lGlocStart + isizeof(int));
 	pgrstrm->SetPositionInFont(lGlocStart);
 	*pfxdBadVersion = ReadVersion(*pgrstrm);
 	pgrstrm->CloseBuffer();
 	if (*pfxdBadVersion > kGlocVersion)
 		return false;
 
-	pgrstrm->OpenBuffer(feat_tbl, isizeof(int));
+	pgrstrm->OpenBuffer(pFeatTbl, isizeof(int));
 	pgrstrm->SetPositionInFont(lFeatStart);
 	*pfxdBadVersion = ReadVersion(*pgrstrm);
 	pgrstrm->CloseBuffer();
@@ -1024,7 +1072,7 @@ bool GrEngine::ReadSilfTable(GrIStream & grstrm, long lTableStart, int iSubTable
 	data16 chwTmp;
 	*pchwMaxGlyphID = grstrm.ReadUShortFromFont();
 
-	//	extra ascent
+	//	extra ascent and descent
 	m_mXAscent = grstrm.ReadShortFromFont();
 	m_mXDescent = grstrm.ReadShortFromFont();
 
@@ -1073,9 +1121,9 @@ bool GrEngine::ReadSilfTable(GrIStream & grstrm, long lTableStart, int iSubTable
 	bTmp = grstrm.ReadByteFromFont();
 	m_chwDirAttr = bTmp;
 
-	//	Sanity checks.
-	if (m_chwPseudoAttr > 100 || m_chwBWAttr > 100 || m_chwDirAttr > 100)
-		return false; // bad table
+	//	Sanity checks--don't bother with these, I don't seem to be able to know what are reasonable values.
+	//if (m_chwPseudoAttr > 200 || m_chwBWAttr > 200 || m_chwDirAttr > 200)
+	//	return false; // bad table
 
 	if (*pfxdSilfVersion >= 0x00020000)
 	{
@@ -1181,7 +1229,7 @@ bool GrEngine::ReadSilfTable(GrIStream & grstrm, long lTableStart, int iSubTable
 		grstrm.SetPositionInFont(lPassBlockPos);
 	else
 		//	Otherwise assume that's where we are!
-		Assert(lPassBlockPos = -1);
+		Assert(lPassBlockPos == -1);
 
 	//	offsets to passes, relative to the start of this subtable;
 	//	note that we read (cPasses + 1) of these
@@ -1196,7 +1244,7 @@ bool GrEngine::ReadSilfTable(GrIStream & grstrm, long lTableStart, int iSubTable
 		grstrm.SetPositionInFont(lPseudosPos);
 	else
 		//	Otherwise assume that's where we are!
-		Assert(lPseudosPos = -1);
+		Assert(lPseudosPos == -1);
 
 	//	number of pseudo-glyphs and search constants
 	short snTmp;
@@ -1293,6 +1341,8 @@ void GrEngine::CreateEmpty()
 
 	//	Feat table
 	m_cfeat = 0;
+	m_rglcidFeatLabelLangs = NULL;
+	m_clcidFeatLabelLangs = 0;
 
 	//	Language table
 	m_langtbl.CreateEmpty();
@@ -1301,8 +1351,8 @@ void GrEngine::CreateEmpty()
 /*----------------------------------------------------------------------------------------------
 	Create the glyph table and fill it in from the font stream.
 ----------------------------------------------------------------------------------------------*/	
-bool GrEngine::ReadGlocAndGlatTables(GrIStream & gloc_strm, long lGlocStart,
-	GrIStream & glat_strm, long lGlatStart,
+bool GrEngine::ReadGlocAndGlatTables(GrIStream & grstrmGloc, long lGlocStart,
+	GrIStream & grstrmGlat, long lGlatStart,
 	int chwGlyphIDMax, int fxdSilfVersion)
 {
 	//	Determine the highest used glyph ID number.
@@ -1319,7 +1369,7 @@ bool GrEngine::ReadGlocAndGlatTables(GrIStream & gloc_strm, long lGlocStart,
 	m_pgtbl->SetNumberOfComponents(m_cComponents);
 	m_pgtbl->SetNumberOfStyles(1);	// for now
 
-	return m_pgtbl->ReadFromFont(gloc_strm, lGlocStart, glat_strm, lGlatStart,
+	return m_pgtbl->ReadFromFont(grstrmGloc, lGlocStart, grstrmGlat, lGlatStart,
 		m_chwBWAttr, m_chwJStretch0, m_cJLevels, m_cnCompPerLig, fxdSilfVersion);
 }
 
@@ -1379,11 +1429,19 @@ bool GrEngine::ReadFeatTable(GrIStream & grstrm, long lTableStart)
 		//	index into name table of UI label
 		int nNameId = grstrm.ReadShortFromFont();
 
+		if (fxdVersion <= 0x00020000 && nID == GrFeature::knLangFeatV2)
+		{
+			// Ignore the obsolete "lang" feature which has ID = 1
+			vnIDs.pop_back();
+			vcfset.pop_back();
+			vnOffsets.pop_back();
+			continue;
+		}
 		AddFeature(nID, nNameId, cfset);
 	}
 
 	//	default feature setting value and name strings;
-	for (ifeat = 0; ifeat < cfeat; ifeat++)
+	for (ifeat = 0; ifeat < m_cfeat; ifeat++)
 	{
 		GrFeature * pfeat = m_rgfeat + ifeat;
 		Assert(pfeat->ID() == vnIDs[ifeat]);
@@ -1452,7 +1510,7 @@ GrResult GrEngine::RunUsingEmpty(Segment * psegNew, Font * pfont,
 	int cComponentsSave = m_cComponents;
 	int cnUserDefnSave = m_cnUserDefn;
 	int cnCompPerLigSave = m_cnCompPerLig;
-	int grfsdcSave = m_grfsdc;
+	unsigned int grfsdcSave = m_grfsdc;
 	gid16 chwLBGlyphIDSave = m_chwLBGlyphID;
 	int cpsdSave = m_cpsd;
 	int dipsdInitSave = m_dipsdInit;
@@ -1543,9 +1601,9 @@ gid16 GrEngine::GetGlyphIDFromUnicode(int nUnicode)
 	//	Otherwise, get the glyph ID from the font's cmap.
 
 	if (m_pCmap_3_10)
-		return TtfUtil::Cmap310Lookup(m_pCmap_3_10, nUnicode);
+		return gid16(TtfUtil::Cmap310Lookup(m_pCmap_3_10, nUnicode));
 	else if (m_pCmap_3_1)
-		return TtfUtil::Cmap31Lookup(m_pCmap_3_1, nUnicode);
+		return gid16(TtfUtil::Cmap31Lookup(m_pCmap_3_1, nUnicode));
 	else
 		return 0;
 }
@@ -1600,7 +1658,7 @@ gid16 GrEngine::MapToPseudo(int nUnicode)
 ----------------------------------------------------------------------------------------------*/
 gid16 GrEngine::ActualGlyphForOutput(gid16 chwGlyphID)
 {
-	gid16 chwActual = GlyphAttrValue(chwGlyphID, m_chwPseudoAttr);
+	gid16 chwActual = gid16(GlyphAttrValue(chwGlyphID, m_chwPseudoAttr));
 	if (chwActual == 0)
 		return chwGlyphID; // not a pseudo, we're already working with the actual glyph ID
 	else
@@ -1615,18 +1673,15 @@ std::wstring GrEngine::StringFromNameTable(int nLangID, int nNameID)
 {
 	std::wstring stuName;
 	stuName.erase();
-	long lOffset = -1;
-	long lSize = -1;
+	size_t lOffset = 0, lSize = 0;
 
 	// The Graphite compiler stores our names in either 
 	// the MS (platform id = 3) Unicode (writing system id = 1) table
 	// or the MS Symbol (writing system id = 0) table. Try MS Unicode first.
 	// lOffset & lSize are in bytes.
-	////if (!TtfUtil::GetNameInfo(m_tbufNameTbl, 3, 1, nLangID, nNameID, lOffset, lSize))
 	// new interface:
 	if (!TtfUtil::GetNameInfo(m_pNameTbl, 3, 1, nLangID, nNameID, lOffset, lSize))
 	{
-		////if (!TtfUtil::GetNameInfo(m_tbufNameTbl, 3, 0, nLangID, nNameID, lOffset, lSize))
 		if (!TtfUtil::GetNameInfo(m_pNameTbl, 3, 0, nLangID, nNameID, lOffset, lSize))
 		{
 			return stuName;
@@ -1635,25 +1690,19 @@ std::wstring GrEngine::StringFromNameTable(int nLangID, int nNameID)
 
 	size_t cchw = (unsigned(lSize) / sizeof(utf16));
 	utf16 * pchwName = new utf16[cchw+1]; // lSize - byte count for Uni str
-//	memcpy(pchwName, (byte *)m_pNameTbl + lOffset, lSize);
-//	pchwName[cchw] = 0;  // zero terminate
-    ////const utf16 *pchwSrcName = reinterpret_cast<const utf16*>(m_tbufNameTbl + lOffset);
 	const utf16 *pchwSrcName = reinterpret_cast<const utf16*>(m_pNameTbl + lOffset);
     std::transform(pchwSrcName, pchwSrcName + cchw, pchwName, std::ptr_fun<utf16,utf16>(lsbf));
 	pchwName[cchw] = 0;  // zero terminate
-//	if (TtfUtil::SwapWString(pchwName, cchw))
-//	{
 	#ifdef _WIN32
-		stuName.assign(pchwName, cchw);
+		stuName.assign((const wchar_t*)pchwName, cchw);
 	#else
 		wchar_t * pchwName32 = new wchar_t[cchw+1]; // lSize - byte count for Uni str
-		for (int i=0; i<=(signed)cchw;i++) {
+		for (int i = 0; i <= signed(cchw); i++) {
 			pchwName32[i] = pchwName[i];
 		}
 		stuName.assign(pchwName32, cchw);
 		delete [] pchwName32;
 	#endif
-//	}
 
 	delete [] pchwName;
 	return stuName;
