@@ -103,7 +103,7 @@ public:
 	//	std::ostream  * pstrmLog,
 	//	OLECHAR * prgchwErrMsg, int cchMaxErrMsg);
 
-	GrResult get_ScriptDirection(int * pgrfsdc, OLECHAR * prgchwErrMsg, int cchMaxErrMsg);
+	GrResult get_ScriptDirection(unsigned int * pgrfsdc, OLECHAR * prgchwErrMsg, int cchMaxErrMsg);
 
 	// DELETE
 	//GrResult MakeJustifiedSegment(Font * pfont, ITextSource * pgts, IGrJustifier * pgj,
@@ -405,14 +405,16 @@ protected:
 									// control file
 
 	GrResult m_resFontRead;		// kresOk if the Graphite font was loaded successfully;
-								// kresFail if the Graphite font could not be found;
+								// kresFail if the font could not be found or basic tables couldn't be read;
+								// kresFalse if the Silf table is not present;
 								// kresUnexpected if the Graphite tables could not be loaded
 
-	GrResult m_resFontValid;	// kresOk if Graphite font & dc font match cmaps
-								// kresUnexpected if dumb rendering
-								// kresFail if dumb rendering with no cmap
-								// kresInvalidArg if the engine is not yet initialized
-
+	GrResult m_resFontValid;	// kresInvalidArg if the engine is not yet initialized
+								// Originally:
+								//	kresOk if Graphite font & dc font match cmaps
+								//	kresUnexpected/kresFalse if dumb rendering
+								//	kresFail if dumb rendering with no cmap
+								// But now it is apparently just = to m_resFontRead otherwise.
 	FontErrorCode m_ferr;
 	int m_fxdBadVersion;
 
@@ -420,7 +422,7 @@ protected:
 							// check sum as a unique identifer
 
 	int m_nScriptTag;		// from the pertinent writing system; currently not used
-	int m_grfsdc;			// supported script directions
+	unsigned int m_grfsdc;	// supported script directions
 
 	bool m_fRightToLeft;	// overall writing-system direction
 
@@ -444,6 +446,9 @@ protected:
 //									// Review: do we need this, or is a linear search adequate?
 	int	m_cfeat;					// number of features present
 
+	short * m_rglcidFeatLabelLangs;	// LCIDs (language IDs) for all the labels in the feature table
+	size_t m_clcidFeatLabelLangs;
+
 	GrLangTable m_langtbl;
 
 	//	are line-breaks relevant at all?
@@ -488,10 +493,12 @@ protected:
 							//		use for Unicode to Glyph ID conversion
 	void * m_pCmap_3_10;
 	byte * m_pCmapTbl;
+	bool m_fCmapTblCopy;
 
 	// for feature names and maybe other strings from font later
 	////TableBuffer	m_tbufNameTbl;	// hold full name table; use Name() method to access
 	byte * m_pNameTbl;
+	bool m_fNameTblCopy;
 
 	bool m_fLogXductn;	// true if we want to log the transduction process
 
@@ -516,7 +523,7 @@ protected:
 		float dxMaxWidth,
 		bool fBacktracking,
 		// for making a justified segment
-		float dxJustifiedWidthJ, SegEnd estJ);
+		bool fJust, float dxJustifiedWidthJ, SegEnd estJ);
 
 	int FindFontLim(ITextSource * pgts, int ichFontMin, int * pnDirDepth);
 
@@ -543,9 +550,23 @@ protected:
 		int chwMaxGlyphID, int fxdSilfVersion);
 	bool ReadFeatTable(GrIStream & grstrm, long lTableStart);
 	bool ReadSillTable(GrIStream & grstrm, long lTableStart);
+	bool SetCmapAndNameTables(Font * pfont);
 
 	void CreateEmpty();
 
+	bool BadFont(FontErrorCode * pferr = NULL)
+	{
+		if (pferr)
+			*pferr = m_ferr;
+		return (m_resFontValid == kresFail || m_resFontRead == kresFail);
+	}
+	bool DumbFallback(FontErrorCode * pferr = NULL)
+	{
+		if (pferr)
+			*pferr = m_ferr;
+		return (m_resFontValid != kresOk || m_resFontRead != kresOk);
+	}
+
 	gid16 MapToPseudo(int nUnicode);
 
 	// NEW interface
@@ -560,10 +581,15 @@ protected:
 	size_t NumberOfSettings_ff(size_t ifeat);
 	int GetFeatureSettingValue_ff(size_t ifeat, size_t ifset);
 	bool GetFeatureSettingLabel_ff(size_t ifeat, size_t ifset, lgid language, utf16 * label);
+	// Feature-label languages:
+	size_t NumberOfFeatLangs_ff();
+	short GetFeatLabelLang_ff(size_t ilang);
 	// Language access for FontFace:
 	size_t NumberOfLanguages_ff();
 	isocode GetLanguageCode_ff(size_t ilang);
 
+	void SetUpFeatLangList();
+
 //:Ignore
 #ifdef OLD_TEST_STUFF
 	//	For test procedures:
