@@ -44,6 +44,7 @@ namespace gr
 ----------------------------------------------------------------------------------------------*/
 GrPass::GrPass(int i)
 	:	m_ipass(i),
+		vnStack(128),
 		m_nMaxRuleContext(0),
 		m_pfsm(NULL),
 		m_nMaxRuleLoop(0),
@@ -57,9 +58,9 @@ GrPass::GrPass(int i)
 		m_prgbConstraintBlock(NULL),
 		m_prgbActionBlock(NULL),
 		m_prgibConstraintDebug(NULL),
-        m_prgibRuleDebug(NULL),
-        m_fCheckRules(false),
-        m_prgfRuleOkay(NULL)
+		m_prgibRuleDebug(NULL),
+		m_fCheckRules(false),
+		m_prgfRuleOkay(NULL)
 {
 }
 
@@ -76,8 +77,8 @@ void PassState::InitForNewSegment(int ipass, int nMaxChunk)
 void PassState::InitializeLogInfo()
 {
 	m_crulrec = 0;
-	memset(m_rgcslotDeletions, 0, 128 * sizeof(int));
-	memset(m_rgfInsertion, 0, 128 * sizeof(bool));
+	std::fill_n(m_rgcslotDeletions, 128, 0);
+	std::fill_n(m_rgfInsertion, 128, false);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -710,7 +711,8 @@ int GrPass::ExtendGlyphIDOutput(GrTableManager * ptman,
 		int nUnicode = pchstrm->NextGet(ptman, &fval, &ichwSegOffset, &cchw);
 		gid16 chwGlyphID = ptman->GetGlyphIDFromUnicode(nUnicode);
 
-		if (nUnicode == knCR || nUnicode == knLF || nUnicode == knLineSep || nUnicode == knParaSep)
+		if (nUnicode == knCR || nUnicode == knLF || nUnicode == knLineSep || nUnicode == knParaSep ||
+			nUnicode == knORC)
 		{
 			//	Hard line-break: we're done.
 			//	Note that we don't include the hard-break character in this segment.
@@ -771,7 +773,8 @@ int GrPass::ExtendFinalOutput(GrTableManager * ptman,
 //	Assert(m_cslotSkipToResync == 0);
 
 	int islotOutputLB = -1;
-	int islotInitWritePos = psstrmOutput->WritePos();
+	//int islotInitWritePos = psstrmOutput->WritePos();
+	int islotNoLbUpTo = psstrmOutput->WritePos();;
 
 	while (true)
 	{
@@ -802,6 +805,7 @@ int GrPass::ExtendFinalOutput(GrTableManager * ptman,
 
 			*pislotLB = -1;
 			pengst->SetExceededSpace();
+			pengst->SetHitHardBreak(false);
 			return kBacktrack;
 		}
 
@@ -856,9 +860,12 @@ int GrPass::ExtendFinalOutput(GrTableManager * ptman,
 		RunRule(ptman, ruln, psstrmInput, psstrmOutput);
 
 		if (fHaveLineBreak)
+		{
 			islotOutputLB =
 				psstrmOutput->FindFinalLineBreak(ptman->LBGlyphID(),
-					islotInitWritePos, psstrmOutput->WritePos());
+					islotNoLbUpTo, psstrmOutput->WritePos());
+			islotNoLbUpTo = psstrmOutput->WritePos();
+		}
 
 		psstrmOutput->CalcIndexOffset(ptman);
 	}
@@ -936,12 +943,16 @@ void GrPass::CheckInputProgress(GrSlotStream * psstrmInput, GrSlotStream * psstr
 			while (!psstrmInput->AtEnd() &&
 				psstrmInput->ReadPosForNextGet() < psstrmInput->ReadPosMax())
 			{
+				RecordHitMaxRuleLoop(psstrmInput->ReadPosForNextGet());
 				psstrmOutput->CopyOneSlotFrom(psstrmInput);
 				fAdvanced = true;
 			}
 			// If that didn't do anything productive, just advance one slot.
 			if (!fAdvanced && !psstrmInput->AtEndOfContext())
+			{
+				RecordHitMaxRuleLoop(psstrmInput->ReadPosForNextGet());
 				psstrmOutput->CopyOneSlotFrom(psstrmInput);
+			}
 
 			m_pzpst->m_nRulesSinceAdvance = 0;
 		}
@@ -1708,7 +1719,7 @@ int GrPass::Unwind(GrTableManager * ptman,
 	if (!psstrmIn->NoReproc())
 		islotIn = min(islotIn, psstrmIn->ReprocMin());
 	psstrmIn->ClearReprocBuffer();
-	int islotOut;
+	int islotOut = 0;
 	if (islotIn < psstrmIn->SlotsSkippedToResync() ||
 		islotIn == 0 ||
 		psstrmIn->ReadPos() == 0)
@@ -1886,25 +1897,16 @@ void GrPosPass::Unattach(GrSlotStream * psstrmIn, int islotIn,
 }
 //:End Ignore
 
+
 /*----------------------------------------------------------------------------------------------
 	Record the fact that the given rule's constraint failed, for the purpose of logging
 	the transduction process.
 ----------------------------------------------------------------------------------------------*/
 void GrPass::RecordRuleFailed(int islot, int irul)
 {
-	m_pzpst->RecordRuleFailed(islot, irul);
+	m_pzpst->RecordRule(islot, irul, false);
 }
 
-void PassState::RecordRuleFailed(int islot, int irul)
-{
-	if (m_crulrec >= 128)
-		return;
-
-	m_rgrulrec[m_crulrec].m_irul = irul;
-	m_rgrulrec[m_crulrec].m_islot = islot;
-	m_rgrulrec[m_crulrec].m_fFired = false;
-	m_crulrec++;
-}
 
 /*----------------------------------------------------------------------------------------------
 	Record the fact that the given rule fired, for the purpose of logging
@@ -1912,10 +1914,36 @@ void PassState::RecordRuleFailed(int islot, int irul)
 ----------------------------------------------------------------------------------------------*/
 void GrPass::RecordRuleFired(int islot, int irul)
 {
-	m_pzpst->RecordRuleFired(islot, irul);
+	m_pzpst->RecordRule(islot, irul, true);
 }
 
-void PassState::RecordRuleFired(int islot, int irul)
+
+/*----------------------------------------------------------------------------------------------
+	Record the fact that we forcibly advanced due to the max-rule-loop.
+----------------------------------------------------------------------------------------------*/
+void GrPass::RecordHitMaxRuleLoop(int islot)
+{
+	m_pzpst->RecordRule(islot, PassState::kHitMaxRuleLoop);
+}
+
+
+/*----------------------------------------------------------------------------------------------
+	Record the fact that we forcibly advanced due to the max-backup.
+
+	In theory it should be possible to call this, but in reality, it is pretty hard to be
+	able to say that there is a rule that would have fired except it was prevent by
+	MaxBackup. So currently this method is not called.
+----------------------------------------------------------------------------------------------*/
+void GrPass::RecordHitMaxBackup(int islot)
+{
+	m_pzpst->RecordRule(islot, PassState::kHitMaxBackup);
+}
+
+
+/*----------------------------------------------------------------------------------------------
+	Record some rule-related activity on a slot.
+----------------------------------------------------------------------------------------------*/
+void PassState::RecordRule(int islot, int irul, bool fFired)
 {
 	if (m_crulrec >= 128)
 		return;
@@ -1923,9 +1951,11 @@ void PassState::RecordRuleFired(int islot, int irul)
 	m_rgrulrec[m_crulrec].m_irul = irul;
 	m_rgrulrec[m_crulrec].m_islot = islot;
 	m_rgrulrec[m_crulrec].m_fFired = true;
+	m_rgrulrec[m_crulrec].m_fFired = fFired;
 	m_crulrec++;
 }
 
+
 /*----------------------------------------------------------------------------------------------
 	Record the fact that a slot was deleted before the given slot (in the output).
 ----------------------------------------------------------------------------------------------*/
