@@ -78,8 +78,7 @@ public:
 	}
 
 	void UnwindLogInfo(int islotIn, int islotOut);
-	void RecordRuleFailed(int islot, int irul);
-	void RecordRuleFired(int islot, int irul);
+	void RecordRule(int islot, int irul, bool fFired = false);
 	void RecordDeletionBefore(int islot);
 	void RecordInsertionAt(int islot);
 #ifdef TRACING
@@ -110,10 +109,15 @@ protected:
 	int m_cslotSkipToResync;
 	bool m_fDidResyncSkip;
 
+	enum {
+		kHitMaxBackup = -1,
+		kHitMaxRuleLoop = -2
+	};
+
 	//	For logging transduction process: record of one rule matched or fired
 	struct RuleRecord
 	{
-		int m_irul;		// rule index
+		int m_irul;		// rule index, or kHitMaxBackup or kHitMaxRuleLoop
 		int m_islot;	// slot index of input stream
 		bool m_fFired;
 	};
@@ -186,6 +190,8 @@ public:
 	//	Destructor:
 	virtual ~GrPass();
 
+	int PassNumber() { return m_ipass; }
+
 	bool ReadFromFont(GrIStream & grstrm, int fxdSilfVersion, int fxdRuleVersion, int nOffset);
 	void InitializeWithNoRules();
 
@@ -351,6 +357,8 @@ public:
 
 	void RecordRuleFailed(int islot, int irul);
 	void RecordRuleFired(int islot, int irul);
+	void RecordHitMaxRuleLoop(int islot);
+	void RecordHitMaxBackup(int islot);
 #ifdef TRACING
 	void LogRulesFiredAndFailed(std::ostream & strmOut, GrSlotStream * psstrmIn);
 	void LogInsertionsAndDeletions(std::ostream & strmOut, GrSlotStream * psstrmOut);
@@ -386,6 +394,7 @@ protected:
 
 	int m_fxdVersion;
 
+	std::vector<int> vnStack;	
 	//	number of items required from previous pass; don't access directly, use the getter
 	//	method, because GrBidiPass overrides to always use 1.
 	int m_nMaxRuleContext;
