@@ -64,8 +64,9 @@ int GrPass::RunCommandCode(GrTableManager * ptman,
 {
 	byte * pbNext = pbStart;
 
-	std::vector<int> vnStack;
+//	std::vector<int> vnStack;
 //	vnStack.EnsureSpace(128);	// make it nice and big
+    vnStack.clear();
 
 	gAssert(fConstraints || islot == 0);
 
