@@ -42,18 +42,24 @@ namespace gr
 	Initialize slots.
 ----------------------------------------------------------------------------------------------*/
 //	standard for pass 0 slots
-void GrSlotState::Initialize(int chw, GrEngine * pgreng,
+void GrSlotState::Initialize(gid16 chw, GrEngine * pgreng,
 	GrFeatureValues fval, int ipass, int ichwSegOffset, int nUnicode)
 {
 	Assert(ipass == 0);
 	m_chwGlyphID = chw;
 	m_chwActual = kInvalidGlyph;
 	m_xysGlyphWidth = -1;
-	m_bStyleIndex = fval.m_nStyleIndex;
-	memset(PUserDefnBuf(), 0, (isizeof(u_intslot) * m_cnUserDefn));
-	memset(PCompRefBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
-	memset(PSlatiBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
-	memcpy(PFeatureBuf(), fval.m_rgnFValues, (isizeof(u_intslot) * m_cnFeat));
+	m_bStyleIndex = byte(fval.m_nStyleIndex);
+	u_intslot nullSlot;
+	nullSlot.pslot = NULL;
+	std::fill_n(PUserDefnBuf(), m_cnUserDefn, nullSlot);
+	std::fill_n(PCompRefBuf(), m_cnCompPerLig, nullSlot);
+	std::fill_n(PSlatiBuf(), m_cnCompPerLig, nullSlot);
+	u_intslot * pFeatBuf = PFeatureBuf();
+	for (size_t i = 0; i < m_cnFeat; i++)
+		pFeatBuf[i].nValue = fval.m_rgnFValues[i];
+	m_ipassFsmCol = -1;
+	m_colFsm = -1;
 
 	m_ipassModified = ipass;
 	m_ichwSegOffset = ichwSegOffset;
@@ -78,15 +84,17 @@ void GrSlotState::Initialize(int chw, GrEngine * pgreng,
 }
 
 //	line-break slots
-void GrSlotState::Initialize(int chw, GrEngine * pgreng,
+void GrSlotState::Initialize(gid16 chw, GrEngine * pgreng,
 	GrSlotState * pslotFeat, int ipass, int ichwSegOffset)
 {
 	m_chwGlyphID = chw;
 	m_chwActual = kInvalidGlyph;
 	m_xysGlyphWidth = -1;
-	memset(PUserDefnBuf(), 0, (isizeof(u_intslot) * m_cnUserDefn));
-	memset(PCompRefBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
-	memset(PSlatiBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
+	u_intslot nullSlot;
+	nullSlot.pslot = NULL;
+	std::fill_n(PUserDefnBuf(), m_cnUserDefn, nullSlot);
+	std::fill_n(PCompRefBuf(), m_cnCompPerLig, nullSlot);
+	std::fill_n(PSlatiBuf(), m_cnCompPerLig, nullSlot);
 	CopyFeaturesFrom(pslotFeat);
 	m_ipassModified = ipass;
 	m_ichwSegOffset = ichwSegOffset;
@@ -94,18 +102,22 @@ void GrSlotState::Initialize(int chw, GrEngine * pgreng,
 	m_vpslotAssoc.clear();
 	pgreng->InitSlot(this);
 	// Caller is responsible for setting m_spsl.
+	m_ipassFsmCol = -1;
+	m_colFsm = -1;
 }
 
 //	for inserting new slots after pass 0 (under-pos and unicode are irrelevant)
-void GrSlotState::Initialize(int chw, GrEngine * pgreng,
+void GrSlotState::Initialize(gid16 chw, GrEngine * pgreng,
 	GrSlotState * pslotFeat, int ipass)
 {
 	m_chwGlyphID = chw;
 	m_chwActual = kInvalidGlyph;
 	m_xysGlyphWidth = -1;
-	memset(PUserDefnBuf(), 0, (isizeof(u_intslot) * m_cnUserDefn));
-	memset(PCompRefBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
-	memset(PSlatiBuf(), 0, (isizeof(u_intslot) * m_cnCompPerLig));
+	u_intslot nullSlot;
+	nullSlot.pslot = NULL;
+	std::fill_n(PUserDefnBuf(), m_cnUserDefn, nullSlot);
+	std::fill_n(PCompRefBuf(), m_cnCompPerLig, nullSlot);
+	std::fill_n(PSlatiBuf(), m_cnCompPerLig, nullSlot);
 	CopyFeaturesFrom(pslotFeat);
 	m_ipassModified = ipass;
 	m_ichwSegOffset = kInvalid;
@@ -113,6 +125,8 @@ void GrSlotState::Initialize(int chw, GrEngine * pgreng,
 	m_vpslotAssoc.clear();
 	pgreng->InitSlot(this);
 	m_spsl = kspslNone;
+	m_ipassFsmCol = -1;
+	m_colFsm = -1;
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -127,6 +141,10 @@ void GrSlotState::InitializeFrom(GrSlotState * pslotOld, int ipass)
 	m_vpslotAssoc.clear();
 	m_vpslotAssoc.push_back(pslotOld);
 
+	// Since we're going on to a new pass, no point in copying these:
+	m_ipassFsmCol = -1;	
+	m_colFsm = -1;
+
 	m_dislotRootFixed = pslotOld->m_dislotRootFixed;
 	m_vdislotAttLeaves.resize(pslotOld->m_vdislotAttLeaves.size());
 	for (size_t i = 0; i < pslotOld->m_vdislotAttLeaves.size(); i++)
@@ -144,7 +162,7 @@ void GrSlotState::CopyFeaturesFrom(GrSlotState * pslotSrc)
 {
 	m_bStyleIndex = pslotSrc->m_bStyleIndex;
 	Assert(m_cnFeat == pslotSrc->m_cnFeat);
-	memcpy(PFeatureBuf(), pslotSrc->PFeatureBuf(), m_cnFeat*sizeof(u_intslot));
+	std::copy(pslotSrc->PFeatureBuf(), pslotSrc->PFeatureBuf() + m_cnFeat, PFeatureBuf());
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -156,10 +174,10 @@ void GrSlotState::CopyFeaturesFrom(GrSlotState * pslotSrc)
 void GrSlotAbstract::CopyFrom(GrSlotState * pslot)
 {
 	u_intslot * pnBufSave = m_prgnVarLenBuf;
-	memcpy(this, pslot, isizeof(GrSlotAbstract));
+	*this = *pslot;
 	m_prgnVarLenBuf = pnBufSave;
 	Assert(m_prgnVarLenBuf);
-	memcpy(m_prgnVarLenBuf, pslot->m_prgnVarLenBuf, (isizeof(u_intslot) * CExtraSpace()));
+	std::copy(pslot->m_prgnVarLenBuf, pslot->m_prgnVarLenBuf + CExtraSpace(), m_prgnVarLenBuf);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -224,7 +242,7 @@ void GrSlotState::EnsureCacheForOutput(GrTableManager * ptman)
 		m_xysGlyphWidth = 0;
 		m_xysAdvX = 0;
 		m_xysAdvY = 0;
-		m_fIsSpace = true;
+		m_bIsSpace = true;
 	}
 	else
 	{
@@ -270,7 +288,8 @@ void GrSlotState::Associate(std::vector<GrSlotState*> & vpslot)
 	//	Set its character styles and features from the associated slot.
 	if (vpslot.size() > 0) // && !m_pslotPrevState
 	{
-		memcpy(PFeatureBuf(), m_vpslotAssoc[0]->PFeatureBuf(), (isizeof(int) * m_cnFeat));
+		std::copy(m_vpslotAssoc[0]->PFeatureBuf(),
+				  m_vpslotAssoc[0]->PFeatureBuf() + m_cnFeat, PFeatureBuf());
 	}
 }
 
@@ -473,7 +492,7 @@ GrSlotState * GrSlotAbstract::CompRefSlot(int i)
 {
 	Assert(i < m_cnCompPerLig);
 	if (i < m_cnCompPerLig)
-		return reinterpret_cast<GrSlotState *>(CompRef(i));
+		return CompRef(i);
 	else
 		return NULL;
 }
@@ -709,15 +728,15 @@ float GrSlotAbstract::GetGlyphMetric(Font * pfont, int nMetricID, gid16 chwGlyph
 		pfont->getGlyphMetrics(chwGlyphID, rectBb, ptAdvances);
 
 		m_xysGlyphX = rectBb.left;
-		m_xysGlyphY = rectBb.bottom;
+		m_xysGlyphY = rectBb.top;
 		m_xysGlyphWidth = (rectBb.right - rectBb.left);
 		m_xysGlyphHeight = (rectBb.top - rectBb.bottom);
 		m_xysAdvX = ptAdvances.x;
 		m_xysAdvY = ptAdvances.y;
 
-		m_fIsSpace = (0 == m_xysGlyphX && 0 == m_xysGlyphY); // should agree with test done in IsSpace() below
+		m_bIsSpace = (0 == m_xysGlyphX && 0 == m_xysGlyphY); // should agree with test done in IsSpace() below
 
-		if (m_fIsSpace)
+		if (m_bIsSpace == 1)
 		{
 			// White space glyph - only case where nGlyphX == nGlyphY == 0
 			// nGlyphWidth & nGlyphHeight are always set to 16 for unknown reasons, so correct.
@@ -777,19 +796,23 @@ int GrSlotState::IsSpace(GrTableManager * ptman)
 	////	ptman->State()->Font()->getGlyphMetrics(gidActual, rectBb, ptAdvances);
 	////}
 
-	GetGlyphMetric(ptman->State()->GetFont(), kgmetBbLeft, gidActual);
+	if (m_bIsSpace == -1)
+		GetGlyphMetric(ptman->State()->GetFont(), kgmetBbLeft, gidActual);
 	// One call is enough to cache the information.
 	//GetGlyphMetric(ptman->State()->Font(), kgmetBbBottom, gidActual);
 
 	// should agree with test done in GetGlyphMetric() above
-	//////m_fIsSpace = (0 == m_xysGlyphX && 0 == m_xysGlyphY);
+	//////m_bIsSpace = (0 == m_xysGlyphX && 0 == m_xysGlyphY);
 
-	return m_fIsSpace;
+	Assert(m_bIsSpace == 0 || m_bIsSpace == 1);
+
+	return m_bIsSpace;
 }
 
 bool GrSlotOutput::IsSpace()
 {
-	return m_fIsSpace;
+	Assert(m_bIsSpace == 0 || m_bIsSpace == 1);
+	return m_bIsSpace;
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -822,6 +845,8 @@ void GrSlotState::HandleModifiedPosition(GrTableManager * ptman,
 
 	if (m_fAttachMod)
 	{
+		//	Attachments.
+
 		//	Note that it doesn't make sense to attach a slot to itself, so if the value of
 		//	the attach.to attribute is 0, that means clear the attachment.
 		GrSlotState * pslotNewRoot = AttachRoot(psstrmOut);
@@ -839,13 +864,13 @@ void GrSlotState::HandleModifiedPosition(GrTableManager * ptman,
 				if (m_mAttachAtX == kNotYetSet && m_nAttachAtGpoint == kNotYetSet)
 					m_mAttachAtX = 0;
 				if (m_mAttachWithX == kNotYetSet && m_nAttachWithGpoint == kNotYetSet)
-					m_mAttachWithX = AdvanceX(ptman);
+					m_mAttachWithX = short(AdvanceX(ptman));
 			}
 			else
 			{
 				//	The root is on the left, the leaf is on the right.
 				if (m_mAttachAtX == kNotYetSet && m_nAttachAtGpoint == kNotYetSet)
-					m_mAttachAtX = pslotNewRoot->AdvanceX(ptman);
+					m_mAttachAtX = short(pslotNewRoot->AdvanceX(ptman));
 				if (m_mAttachWithX == kNotYetSet && m_nAttachWithGpoint == kNotYetSet)
 					m_mAttachWithX = 0;
 			}
@@ -853,6 +878,7 @@ void GrSlotState::HandleModifiedPosition(GrTableManager * ptman,
 	}
 	else
 	{
+		// Shifting, or changing the advance width.
 		Assert(m_fShiftMod);
 		EnsureLocalAttachmentTree(ptman, psstrmIn, psstrmOut, islotThis);
 		ZapMetricsAndPositionDownToBase(psstrmOut);
@@ -882,6 +908,8 @@ void GrSlotState::EnsureLocalAttachmentTree(GrTableManager * ptman,
 	{
 		GrSlotState * pslotRoot = SlotAtOffset(psstrmOut, m_dislotRootFixed);
 		psstrmOut->EnsureLocalCopy(ptman, pslotRoot, psstrmIn);
+		pslotRoot = SlotAtOffset(psstrmOut, m_dislotRootFixed); // get the new one!
+		pslotRoot->EnsureLocalAttachmentTree(ptman, psstrmIn, psstrmOut, islotThis + m_dislotRootFixed);
 	}
 	for (size_t islot = 0; islot < m_vdislotAttLeaves.size(); islot++)
 	{
@@ -1045,9 +1073,9 @@ void GrSlotState::InitLeafMetrics(GrTableManager * ptman, GrSlotState * pslotRoo
 
 	if (IsBase())
 	{
-		//	These two are actually meaningless numbers, for the base, but set them
-		//	to something reasonable.
-		m_xsOffsetX = xsShiftX;
+		//	The x-value below has to be zero because the shift is incorporated into
+		//	m_xsRootShiftX. (m_ysRootShiftY, on the other hand, isn't used anywhere.)
+		m_xsOffsetX = 0;
 		m_ysOffsetY = ysShiftY;
 
 		m_xsRootShiftX = xsShiftX;
@@ -1359,9 +1387,10 @@ void GrSlotOutput::ExactCopyFrom(GrSlotOutput * pslout, u_intslot * pnVarLenBuf,
 {
 	// The chunk of the object from GrSlotAbstract can be copied exactly,
 	// except for the variable-length buffer.
-	memcpy(this, pslout, sizeof(GrSlotAbstract));
+	*this = *pslout;
 	m_prgnVarLenBuf = pnVarLenBuf;
-	memcpy(m_prgnVarLenBuf, pslout->m_prgnVarLenBuf, cnExtraPerSlot * sizeof(u_intslot));
+	std::copy(pslout->m_prgnVarLenBuf, pslout->m_prgnVarLenBuf + cnExtraPerSlot,
+				m_prgnVarLenBuf);
 
 	// Now copy the stuff specific to GrSlotOutput.
 	m_ichwBeforeAssoc = pslout->m_ichwBeforeAssoc;
