@@ -73,6 +73,8 @@ public:
 		m_srAttachTo = 0;
 		m_nAttachLevel = 0;
 
+		m_bIsSpace = -1; // unknown
+
 		m_mAttachAtX = kNotYetSet;
 		m_mAttachAtY = 0;
 		m_nAttachAtGpoint = kNotYetSet;
@@ -97,9 +99,9 @@ public:
 
 		m_islotPosPass = kNotYetSet;
 
-		m_cnUserDefn = cnUserDefn;
-		m_cnCompPerLig = cnCompPerLig;
-		m_cnFeat = cnFeat;
+		m_cnUserDefn = byte(cnUserDefn);
+		m_cnCompPerLig = byte(cnCompPerLig);
+		m_cnFeat = byte(cnFeat);
 		m_prgnVarLenBuf = pnBuf;
 
 		m_fAdvXSet = false; // for transduction logging
@@ -119,8 +121,9 @@ public:
 	void GetFeatureValues(GrFeatureValues * pfval)
 	{
 		pfval->m_nStyleIndex = m_bStyleIndex;
-		memset(pfval->m_rgnFValues, 0, kMaxFeatures);
-		memcpy(pfval->m_rgnFValues, PFeatureBuf(), m_cnFeat);
+		std::fill(pfval->m_rgnFValues, pfval->m_rgnFValues + kMaxFeatures, 0);
+		for (size_t i = 0; i < m_cnFeat; i++)
+			pfval->m_rgnFValues[i] = PFeatureBuf()[i].nValue;
 	}
 
 	int RawAdvanceX()			{ return m_mAdvanceX; }
@@ -228,12 +231,12 @@ public:
 	int UserDefn(int slati)
 	{
 		Assert(slati < m_cnUserDefn);
-		return m_prgnVarLenBuf[slati].integer;
+		return m_prgnVarLenBuf[slati].nValue;
 	}
 	void SetUserDefn(int slati, int nVal)
 	{
 		Assert(slati < m_cnUserDefn);
-		m_prgnVarLenBuf[slati].integer = nVal;
+		m_prgnVarLenBuf[slati].nValue = nVal;
 	}
 	u_intslot * PUserDefnBuf()
 	{
@@ -243,12 +246,12 @@ public:
 	GrSlotState * CompRef(int slati)
 	{
 		Assert(slati < m_cnCompPerLig);
-		return m_prgnVarLenBuf[m_cnUserDefn + slati].slotstate;
+		return m_prgnVarLenBuf[m_cnUserDefn + slati].pslot;
 	}
 	void SetCompRef(int slati, GrSlotState * pvSlot)
 	{
 		Assert(slati < m_cnCompPerLig);
-		m_prgnVarLenBuf[m_cnUserDefn + slati].slotstate = pvSlot;
+		m_prgnVarLenBuf[m_cnUserDefn + slati].pslot = pvSlot;
 	}
 	u_intslot * PCompRefBuf()
 	{
@@ -258,12 +261,12 @@ public:
 	int Slati(int i)
 	{
 		Assert(i < m_cnCompPerLig);
-		return m_prgnVarLenBuf[m_cnUserDefn + m_cnCompPerLig + i].integer;
+		return m_prgnVarLenBuf[m_cnUserDefn + m_cnCompPerLig + i].nValue;
 	}
 	void SetSlati(int i, int n)
 	{
 		Assert(i < m_cnCompPerLig);
-		m_prgnVarLenBuf[m_cnUserDefn + m_cnCompPerLig + i].integer = n;
+		m_prgnVarLenBuf[m_cnUserDefn + m_cnCompPerLig + i].nValue = n;
 	}
 	u_intslot * PSlatiBuf()
 	{
@@ -273,7 +276,7 @@ public:
 	int FeatureValue(int i)
 	{
 		Assert(i < m_cnFeat);
-		return m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + i].integer;
+		return m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + i].nValue;
 	}
 
 	u_intslot * PFeatureBuf()
@@ -358,7 +361,7 @@ protected:
 
 	bool m_fHasComponents;	// default = false
 
-	bool m_fIsSpace;
+	sdata8 m_bIsSpace;	// 0 = false, 1 = true, -1 = unknown
 
 	byte m_bStyleIndex;
 
@@ -422,11 +425,11 @@ public:
 	{
 	}
 
-	void Initialize(int chw, GrEngine *, GrFeatureValues fval,
+	void Initialize(gid16 chw, GrEngine *, GrFeatureValues fval,
 		int ipass, int ichwSegOffset, int nUnicode = -1);
-	void Initialize(int chw, GrEngine *, GrSlotState * pslotFeat,
+	void Initialize(gid16 chw, GrEngine *, GrSlotState * pslotFeat,
 		int ipass, int ichwSegOffset);
-	void Initialize(int chw, GrEngine *, GrSlotState * pslotFeat,
+	void Initialize(gid16 chw, GrEngine *, GrSlotState * pslotFeat,
 		int ipass);
 
 	void InitializeFrom(GrSlotState * pslot, int ipass);
@@ -441,6 +444,8 @@ public:
 		m_chwGlyphID = chw;
 		m_chwActual = kInvalidGlyph;
 		m_xysGlyphWidth = -1; // indicate glyph metrics are invalid
+		m_ipassFsmCol = -1;
+		m_colFsm = -1;
 	}
 	void Associate(GrSlotState *);
 	void Associate(GrSlotState *, GrSlotState *);
@@ -544,7 +549,7 @@ public:
 	{
 		if (m_mAdvanceX == kNotYetSet)
 			//	Initialize it from the glyph metric (adjusted for hinting).
-			m_mAdvanceX = GlyphMetricEmUnits(ptman, kgmetAdvWidth);
+			m_mAdvanceX = short(GlyphMetricEmUnits(ptman, kgmetAdvWidth));
 		return m_mAdvanceX;
 	}
 
@@ -552,7 +557,7 @@ public:
 	{
 		if (m_mAdvanceY == kNotYetSet)
 			//	Initialize it from the glyph metric (adjusted for hinting).
-			m_mAdvanceY = GlyphMetricEmUnits(ptman, kgmetAdvHeight);
+			m_mAdvanceY = short(GlyphMetricEmUnits(ptman, kgmetAdvHeight));
 		return m_mAdvanceY;
 	}
 
@@ -565,7 +570,7 @@ public:
 			if (m_srAttachTo == 0)
 				return 0;
 			else
-				m_mAttachAtX = AttachRoot(psstrm)->AdvanceX(ptman); // attach on the right
+				m_mAttachAtX = short(AttachRoot(psstrm)->AdvanceX(ptman)); // attach on the right
 		}
 		return m_mAttachAtX;
 	}
@@ -609,53 +614,53 @@ public:
 
 	void SetAdvanceX(int mVal)
 	{
-		Assert(mVal < 0xFFFF); m_mAdvanceX = mVal & 0xFFFF; m_fShiftMod = true;
+		Assert(mVal < 0xFFFF); m_mAdvanceX = short(mVal & 0xFFFF); m_fShiftMod = true;
 		m_fAdvXSet = true;	// for transduction logging
 	}
 	void SetAdvanceY(int mVal)
 	{
-		Assert(mVal < 0xFFFF); m_mAdvanceY = mVal & 0xFFFF; m_fShiftMod = true;
+		Assert(mVal < 0xFFFF); m_mAdvanceY = short(mVal & 0xFFFF); m_fShiftMod = true;
 		m_fAdvYSet = true;	// for transduction logging
 	}
 	void SetShiftX(int mVal) {
-		Assert(mVal < 0xFFFF); m_mShiftX = mVal & 0xFFFF; m_fShiftMod = true; }
+		Assert(mVal < 0xFFFF); m_mShiftX = short(mVal & 0xFFFF); m_fShiftMod = true; }
 	void SetShiftY(int mVal) {
-		Assert(mVal < 0xFFFF); m_mShiftY = mVal & 0xFFFF; m_fShiftMod = true; }
+		Assert(mVal < 0xFFFF); m_mShiftY = short(mVal & 0xFFFF); m_fShiftMod = true; }
 
 	void SetAttachTo(int srVal)
 	{
 		Assert(srVal < 0xFFFF);
-		m_srAttachTo = srVal & 0xFFFF;
+		m_srAttachTo = short(srVal & 0xFFFF);
 		m_fAttachMod = true;
 	}
 	void SetAttachLevel(int nVal) {
-		Assert(nVal < 0xFFFF); m_nAttachLevel = nVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(nVal < 0xFFFF); m_nAttachLevel = short(nVal & 0xFFFF); m_fAttachMod = true; }
 
 	void SetAttachAtX(int mVal)
 	{
 		Assert(mVal < 0xFFFF);
-		m_mAttachAtX = mVal & 0xFFFF;
+		m_mAttachAtX = short(mVal) & 0xFFFF;
 		m_fAttachMod = true;
 	}
 	void SetAttachAtY(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachAtY = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachAtY = short(mVal & 0xFFFF); m_fAttachMod = true; }
 	void SetAttachAtGpoint(int nVal) {
-		m_nAttachAtGpoint = nVal; m_fAttachMod = true; }
+		m_nAttachAtGpoint = short(nVal); m_fAttachMod = true; }
 	void SetAttachAtXOffset(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachAtXOffset = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachAtXOffset = short(mVal & 0xFFFF); m_fAttachMod = true; }
 	void SetAttachAtYOffset(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachAtYOffset = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachAtYOffset = short(mVal & 0xFFFF); m_fAttachMod = true; }
 
 	void SetAttachWithX(int mVal)	{
-		Assert(mVal < 0xFFFF); m_mAttachWithX = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachWithX = short(mVal & 0xFFFF); m_fAttachMod = true; }
 	void SetAttachWithY(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachWithY = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachWithY = short(mVal & 0xFFFF); m_fAttachMod = true; }
 	void SetAttachWithGpoint(int nVal) {
-		m_nAttachWithGpoint = nVal; m_fAttachMod = true; }
+		m_nAttachWithGpoint = short(nVal); m_fAttachMod = true; }
 	void SetAttachWithXOffset(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachWithXOffset = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachWithXOffset = short(mVal & 0xFFFF); m_fAttachMod = true; }
 	void SetAttachWithYOffset(int mVal) {
-		Assert(mVal < 0xFFFF); m_mAttachWithYOffset = mVal & 0xFFFF; m_fAttachMod = true; }
+		Assert(mVal < 0xFFFF); m_mAttachWithYOffset = short(mVal & 0xFFFF); m_fAttachMod = true; }
 
 	void SetCompRefSlot(GrTableManager * ptman, int i, GrSlotState * pslotComp);
 
@@ -665,18 +670,18 @@ public:
 	void SetDirectionality(DirCode dirc)	{ m_dirc = dirc; }
 	void SetDirLevel(int n)					{ m_nDirLevel = n; }
 
-	void SetMeasureSol(int mVal)			{ m_mMeasureSol = mVal; }
-	void SetMeasureEol(int mVal)			{ m_mMeasureEol = mVal; }
+	void SetMeasureSol(int mVal)			{ m_mMeasureSol = short(mVal); }
+	void SetMeasureEol(int mVal)			{ m_mMeasureEol = short(mVal); }
 
-	void SetJStretch(int mVal)				{ m_mJStretch0 = mVal; }
-	void SetJShrink(int mVal)				{ m_mJShrink0 = mVal; }
-	void SetJStep(int mVal)					{ m_mJStep0 = mVal; }
-	void SetJWeight(int nVal)				{ m_nJWeight0 = nVal; }
+	void SetJStretch(int mVal)				{ m_mJStretch0 = short(mVal); }
+	void SetJShrink(int mVal)				{ m_mJShrink0 = short(mVal); }
+	void SetJStep(int mVal)					{ m_mJStep0 = short(mVal); }
+	void SetJWeight(int nVal)				{ m_nJWeight0 = byte(nVal); }
 	void SetJWidth(int mVal)				{ m_mJWidth0 = mVal; }
 	void AddJWidthToAdvance(GrTableManager * ptman)
 	{
 		// Don't change m_fShiftMod.
-		m_mAdvanceX = m_mJWidth0 + AdvanceX(ptman); // make sure it is calculated
+		m_mAdvanceX = short(m_mJWidth0 + AdvanceX(ptman)); // make sure it is calculated
 		m_mJWidth0 = 0;
 		m_fAdvXSet = true;	// for transduction logging
 	}
@@ -861,6 +866,15 @@ public:
 	bool IsLrm();
 	bool IsRlm();
 
+	//	Cache of FSM column information for the most recent pass:
+	int FsmColumn() { return m_colFsm; }
+	int PassNumberForColumn() { return m_ipassFsmCol; }
+	void CacheFsmColumn(int ipass, int col)
+	{
+		m_colFsm = col;
+		m_ipassFsmCol = ipass;
+	}
+
 	//	For transduction logging:
 #ifdef TRACING
 	void SlotAttrsModified(bool * rgfMods, bool fPreJust, int * pccomp, int * pcassoc);
@@ -895,6 +909,8 @@ protected:
 	DirCode m_dircProc;		// directionality as processed in bidi algorithm
 	bool m_fDirProcessed;
 
+	int m_colFsm;			// which FSM column this glyph corresponds to...
+	int m_ipassFsmCol;		// ...for the most recent pass
 
 	//	Attachment and metrics
 
@@ -1050,7 +1066,7 @@ public:
 			//	Must be kept consistent with UserDefn() and CompRef() methods inherited
 			//	from GrSlotAbstract. The component indices come after the above variable-
 			//	length buffers.
-			m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat + m_cComponents].integer
+			m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat + m_cComponents].nValue
 				= ichw;
 
 			// OBSOLETE comment:
@@ -1060,7 +1076,7 @@ public:
 			//	characters. This buffer will then hold [0,2].
 
 			m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat
-				+ m_cnCompPerLig + m_cComponents].integer
+				+ m_cnCompPerLig + m_cComponents].nValue
 				= slati;
 
 //			Assert(iComp >= m_cComponents); // because we process them in order they are defined
@@ -1081,12 +1097,12 @@ public:
 		//	Must be kept consistent with UserDefn() and CompRef() methods inherited
 		//	from GrSlotAbstract. The component indices come after the above variable-
 		//	length buffers.
-		return m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat + iComp].integer;
+		return m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat + iComp].nValue;
 	}
 	int ComponentId(int iComp)
 	{
 		return m_prgnVarLenBuf[m_cnUserDefn + (m_cnCompPerLig * 2) + m_cnFeat
-			+ m_cnCompPerLig + iComp].integer;
+			+ m_cnCompPerLig + iComp].nValue;
 	}
 
 	//	Given the index of the component used by the rules, indicate the
