@@ -495,7 +495,8 @@ int GrSlotStream::DirLevelRange(EngineState * pengst, int islotStart, int nTopDi
 			if (fIncludeLast)
 			{
 				// Include in the range to reverse the PDF or the char after the LRM/RLM.
-				for (size_t i = vislotStops.size() - 1; i >= 0 && vislotStops[i] == islot-1; i--)
+				// (Note: don't use a size_t below: we are decrementing below zero.)
+				for (int i = signed(vislotStops.size()) - 1; i >= 0 && vislotStops[i] == islot-1; i--)
 				{
 					vislotStops[i] = islot;
 				}
@@ -1119,7 +1120,7 @@ bool RightToLeftDir(DirCode dirc)
 void GrSlotStream::CopyOneSlotFrom(GrSlotStream * psstrmPrev)
 {
 #ifdef _DEBUG
-	gid16 chw = psstrmPrev->Peek()->GlyphID();
+	gid16 chw; chw = psstrmPrev->Peek()->GlyphID();
 #endif
 
 	// If we are exactly at the segment boundary, pass the information on to this stream.
@@ -1964,8 +1965,13 @@ int GrSlotStream::ChunkInNextMin(int islot)
 int GrSlotStream::ChunkInNextLim(int islot)
 {
 	int islotRet = islot + 1;
+	if (islotRet == m_islotReadPos)
+		return islot;
 	while (m_vislotNextChunkMap[islotRet] == -1 && islotRet < m_islotReadPos)
+	{
 		++islotRet;
+		Assert(islotRet < signed(m_vislotNextChunkMap.size()));
+	}
 	return islotRet;
 }
 
