@@ -347,7 +347,7 @@ void GrTableManager::Run(Segment * psegNew, Font * pfont,
 	m_fLogging = (pstrmLog != NULL);
 
 	int cbPrev;
-	byte * pbPrevSegDat;
+	byte * pbPrevSegDat = NULL;
 	if (psegPrev)
 		cbPrev = psegPrev->NextSegInitBuffer(&pbPrevSegDat);
 	else if (psegInit)
@@ -493,7 +493,7 @@ void GrTableManager::Run(Segment * psegNew, Font * pfont,
 			{
 				// Nothing will fit. Initialize the new segment just enough so that
 				// we can delete it.
-				InitSegmentToDelete(psegNew);
+				InitSegmentToDelete(psegNew, pfont, pchstrm);
 				return;
 			}
 		}
@@ -610,6 +610,17 @@ void GrTableManager::Run(Segment * psegNew, Font * pfont,
 	{
 		return;
 	}
+	else if (psegNew->segmentTermination() == kestHardBreak && 
+		psegNew->startCharacter() == psegNew->stopCharacter())
+	{
+		// Empty segment caused by a hard-break.
+		m_engst.m_lbPrevEnd = klbNoBreak;
+		psegNew->SetWidths(0, 0);
+		psegNew->SetUpOutputArrays(pfont, this, NULL, 0, 0, 0, twsh,
+			fParaRtl, nDirDepth, true);
+		psegNew->SetLayout(layout);
+		return;
+	}
 
 	psegNew->RecordInitializationForThisSeg(cbPrev, pbPrevSegDat);
 
@@ -709,15 +720,24 @@ void GrTableManager::InitNewSegment(Segment * psegNew,
 
 	if (ichwMin >= ichwLim)
 	{
-		// Invalid empty segment.
-		Assert(ichwMin == ichwLim);
-		InitSegmentToDelete(psegNew);
+		if (est == kestHardBreak)
+		{
+			// Empty segment cause by a hard-break.
+			InitSegmentAsEmpty(psegNew, pfont, pchstrm, fStartLine, fEndLine);
+			psegNew->FixTermination(est);
+		}
+		else
+		{
+			// Invalid empty segment.
+			Assert(ichwMin == ichwLim);
+			InitSegmentToDelete(psegNew, pfont, pchstrm);
+		}
 		return;
 	}
 
 	Assert(psegNew);
 
-	psegNew->Initialize(pchstrm->String(),
+	psegNew->Initialize(pchstrm->TextSrc(),
 		ichwMin, ichwLim,
 		lbStart, lbEnd, est, fStartLine, fEndLine, m_pgreng->RightToLeft());
 
@@ -741,10 +761,18 @@ void GrTableManager::InitNewSegment(Segment * psegNew,
 	couldn't find a valid break point. Initialize it just enough so that it can be
 	deleted by the client.
 ----------------------------------------------------------------------------------------------*/
-void GrTableManager::InitSegmentToDelete(Segment * psegNew)
+void GrTableManager::InitSegmentToDelete(Segment * psegNew, Font * pfont,
+	GrCharStream * pchstrm)
 {
-	psegNew->Initialize(NULL, 0, 0, klbClipBreak, klbClipBreak, kestNothingFit,
-		false, false, false);
+	psegNew->Initialize(pchstrm->TextSrc(), 0, 0,
+		klbClipBreak, klbClipBreak, kestNothingFit, false, false,
+		m_pgreng->RightToLeft());
+
+	psegNew->SetEngine(m_pgreng);
+	psegNew->SetFont(pfont);
+	psegNew->SetJustifier(NULL); // can't justify an empty segment
+	psegNew->SetFaceName(m_pgreng->FaceName(), m_pgreng->BaseFaceName());
+	psegNew->SetPreContext(0);
 }
 
 /*----------------------------------------------------------------------------------------------
@@ -857,7 +885,7 @@ void GrTableManager::InitSegmentAsEmpty(Segment * psegNew, Font * pfont,
 
 	Assert(psegNew);
 
-	psegNew->Initialize(pchstrm->String(), ichwMin, ichwLim,
+	psegNew->Initialize(pchstrm->TextSrc(), ichwMin, ichwLim,
 		lbStart, lbEnd, kestNoMore, fStartLine, fEndLine, m_pgreng->RightToLeft());
 
 	psegNew->SetEngine(m_pgreng);
@@ -873,7 +901,7 @@ void GrTableManager::InitSegmentAsEmpty(Segment * psegNew, Font * pfont,
 	if (pbNextSegDat)
 	{
 		byte * pb = pbNextSegDat;
-		*pb++ = lbEnd;
+		*pb++ = byte(lbEnd);
 		*pb++ = kdircNeutral;
 		*pb++ = kdircNeutral;
 		*pb++ = 0;
@@ -965,7 +993,7 @@ bool GrTableManager::Backtrack(int * pislotPrevBreak,
 	Assert(*plbMin <= lbMax);
 	GrSlotStream * psstrmLB = OutputStream(m_cpassLB);
 	islotStartTry = min(islotStartTry, psstrmLB->WritePos() - 1);
-	int islotNewBreak;
+	int islotNewBreak = -1;
 	while (lb <= lbMax)
 	{
 		LineBrk lbNextToTry;
@@ -1132,7 +1160,7 @@ void GrTableManager::InitializeForNextSeg(Segment * pseg,
 	if (islotSurfaceLim == -1)
 		islotSurfaceLim = OutputStream(m_cpass - 1)->FinalSegLim();
 
-	*pb++ = (byte)lbEnd;
+	*pb++ = byte(lbEnd);
 
 	//	Find previous strong and terminator directionality codes.
 
@@ -1152,8 +1180,8 @@ void GrTableManager::InitializeForNextSeg(Segment * pseg,
 			break;
 		}
 	}
-	*pb++ = (byte)dircStrong;
-	*pb++ = (byte)dircTerm;
+	*pb++ = byte(dircStrong);
+	*pb++ = byte(dircTerm);
 
 	//	Record how much of each pass to redo in the next segment.
 	//	The basic algorithm is described in the "WR Data Transform Engine" document.
@@ -1178,8 +1206,8 @@ void GrTableManager::InitializeForNextSeg(Segment * pseg,
 		if (ipass >= m_cpassLB && m_engst.m_fInitialLB)
 			islotBackupMin++; // how far we can back up--not before the start of this segment
 
-		int islotPrevRS; // corresponding slot in the previous stream for rule-start position
-		int islotPrevPC; // same for pre-context position
+		int islotPrevRS = 0; // corresponding slot in the previous stream for rule-start position
+		int islotPrevPC = 0; // same for pre-context position
 
 		//	Initially we have to include at least this many prior slots:
 		int cslotSkipOffset = 0;
@@ -1303,10 +1331,11 @@ LBackupPC:
 	}
 
 	Assert(cslotPreLB >= 0);
+	Assert(cslotPreLB < 0xFF);
 
-	*pb++ = cslotPreLB;
+	*pb++ = byte(cslotPreLB);
 	for (ipass = 0; ipass < m_cpass; ipass++)
-		*pb++ = vcslotSkipOffsets[ipass];
+		*pb++ = byte(vcslotSkipOffsets[ipass]);
 
 	*pcbNextSegDat = 4 + m_cpass;
 
@@ -1703,7 +1732,7 @@ GrTableManager * EngineState::TableManager()
 
 //	standard for pass 0 slots
 void EngineState::NewSlot(
-	int gID, GrFeatureValues fval, int ipass, int ichwSegOffset, int nUnicode,
+	gid16 gID, GrFeatureValues fval, int ipass, int ichwSegOffset, int nUnicode,
 	GrSlotState ** ppslotRet)
 {
 	NextSlot(ppslotRet);
@@ -1712,7 +1741,7 @@ void EngineState::NewSlot(
 
 //	line-break slots
 void EngineState::NewSlot(
-	int gID, GrSlotState * pslotFeat, int ipass, int ichwSegOffset,
+	gid16 gID, GrSlotState * pslotFeat, int ipass, int ichwSegOffset,
 	GrSlotState ** ppslotRet)
 {
 	NextSlot(ppslotRet);
@@ -1721,7 +1750,7 @@ void EngineState::NewSlot(
 
 //	for inserting new slots after pass 0 (under-pos and unicode are irrelevant)
 void EngineState::NewSlot(
-	int gID, GrSlotState * pslotFeat, int ipass,
+	gid16 gID, GrSlotState * pslotFeat, int ipass,
 	GrSlotState ** ppslotRet)
 {
 	NextSlot(ppslotRet);
@@ -2460,7 +2489,7 @@ void GrTableManager::CalcPositionsUpTo(int ipass, GrSlotState * pslotLast,
 				//	Only do this for slots that have actually been processed by the final
 				//	pass, because others may have intermediate values that may be changed
 				//	later. 
-				m_engst.m_islotPosNext = pslotLastBase->PosPassIndex() + psstrm->IndexOffset() + 1;
+				m_engst.m_islotPosNext = pslot->PosPassIndex() + psstrm->IndexOffset() + 1;
 				m_engst.m_xsPosXNext = xs;
 				m_engst.m_ysPosYNext = ys;
 				m_engst.m_dxsTotWidthSoFar = *pxsWidth;
